#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ArVizInteractive.h"

#include <godot_cpp/classes/display_server.hpp>

namespace avi::app
{
    std::string ArVizInteractive::storageName;
    std::string ArVizInteractive::defaultName;

    std::string
    ArVizInteractive::getDefaultName() const
    {
        return defaultName;
    }

    ArVizInteractive*
    ArVizInteractive::Initialize(avi::app::ApplicationSettings& settings)
    {
        storageName = settings.getStorageIdentifier();

        IceUtil::Handle<ArVizInteractiveApp> app(new ArVizInteractiveApp());
        armarx::Application::setInstance(app);

        app->setName(std::string("arviz_godot") + "-" + IceUtil::generateUUID());

        std::thread thread(
            [app]()
            {
                std::array<const char*, 2> arguments = {"arviz_godot", ""};
                int count = 1;

                app->mutex.lock();
                app->main(count, arguments.data());
            });

        std::unique_lock<std::mutex> lock(app->mutex);
        app->condition.wait(lock, [app]() { return app->setupDone; });

        app->thread = std::move(thread);
        return app->component.get();
    }

    bool
    ArVizInteractive::isConnected() const
    {
        return connected;
    }

    armarx::viz::StorageInterfacePrx
    ArVizInteractive::getStorageInterface() const
    {
        return storage;
    }

    armarx::PropertyDefinitionsPtr
    ArVizInteractive::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());
        return def;
    }

    void
    ArVizInteractive::onInitComponent()
    {
        usingProxy(storageName);
    }

    void
    ArVizInteractive::onConnectComponent()
    {
        getProxy(storage, storageName);

        connected = true;
        ARMARX_INFO << "ArVizInteractive connected" << std::endl;
    }

    void
    ArVizInteractive::onDisconnectComponent()
    {
        connected = false;
        ARMARX_INFO << "ArVizInteractive disconnected" << std::endl;
    }

    void
    ArVizInteractive::onExitComponent()
    {
    }

    void
    ArVizInteractive::close()
    {
        std::lock_guard<std::mutex> lock(closedMutex);
        closed = true;
        closedCondition.notify_all();
    }

    void
    ArVizInteractive::ArVizInteractiveApp::setup(
        const armarx::ManagedIceObjectRegistryInterfacePtr& registry,
        Ice::PropertiesPtr properties)
    {
        std::string namePrefix = "arviz_godot";
        size_t tryCount = 0;
        bool done = false;

        while (not done)
        {
            try
            {
                defaultName =
                    tryCount > 0 ? namePrefix + " (" + std::to_string(tryCount) + ")" : namePrefix;
                component = armarx::Component::create<avi::app::ArVizInteractive>(properties);

                registry->addObject(component);
                done = true;

                godot::String title(defaultName.c_str());
                godot::DisplayServer::get_singleton()->window_set_title(title);
            }
            catch (Ice::AlreadyRegisteredException& e)
            {
                tryCount++;
            }
        }

        component->app = this;
        setupDone = true;

        condition.notify_one();
        mutex.unlock();
    }

    int
    ArVizInteractive::ArVizInteractiveApp::exec(const armarx::ArmarXManagerPtr& armarXManager)
    {
        std::unique_lock<std::mutex> lock(component->closedMutex);
        component->closedCondition.wait(lock, [this]() { return component->closed; });

        armarXManager->shutdown();

        return 0;
    }
} // namespace avi::app
