#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "KeyInputEventDispatch.h"

#include <godot_cpp/classes/input_event.hpp>

void
avi::app::KeyInputEventDispatch::_bind_methods()
{
    DETOOLIFY;

}

avi::app::KeyInputEventDispatch::KeyInputEventDispatch()
{
    DETOOLIFY;
    pressedEvents.push_back({"avi_camera_view_switch_projection", "VIEW_CAMERA_SWITCH_PROJECTION"});
    pressedEvents.push_back({"avi_camera_view_reset", "VIEW_CAMERA_RESET"});
    pressedEvents.push_back({"avi_camera_focus_scene", "VIEW_CAMERA_FOCUS_SCENE"});
    pressedEvents.push_back({"avi_camera_focus_point", "VIEW_CAMERA_FOCUS_POINT"});
    pressedEvents.push_back({"avi_lock_mouse", "VIEW_CAMERA_LOCK"});
    pressedEvents.push_back({"avi_lock_escape", "VIEW_CAMERA_ESCAPE_LOCK"});

    pressedEvents.push_back({"avi_hide_ui", "HIDE_UI"});

    pressedEvents.push_back({"avi_screenshot", "TAKE_SCREENSHOT"});
    pressedEvents.push_back({"avi_video", "RECORD_VIDEO"});

    pressedEvents.push_back({"avi_manipulator_none", "MANIPULATOR_NONE"});
    pressedEvents.push_back({"avi_manipulator_move", "MANIPULATOR_MOVE"});
    pressedEvents.push_back({"avi_manipulator_rotate", "MANIPULATOR_ROTATE"});
    pressedEvents.push_back({"avi_manipulator_scale", "MANIPULATOR_SCALE"});

    pressedEvents.push_back({"avi_camera_technical", "CAMERA_TECHNICAL"});
    pressedEvents.push_back({"avi_camera_playful", "CAMERA_PLAYFUL"});

    pressedEvents.push_back({"avi_camera_orbit_negative_horizontal", "ORBIT_NEGATIVE_HORIZONTAL"});
    pressedEvents.push_back({"avi_camera_orbit_positive_horizontal", "ORBIT_POSITIVE_HORIZONTAL"});
    pressedEvents.push_back({"avi_camera_orbit_negative_vertical", "ORBIT_NEGATIVE_VERTICAL"});
    pressedEvents.push_back({"avi_camera_orbit_positive_vertical", "ORBIT_POSITIVE_VERTICAL"});

    pressedEvents.push_back({"place_on_x_numpad", "PLACE_ON_X_NUMPAD"});
    pressedEvents.push_back({"place_on_y_numpad", "PLACE_ON_Y_NUMPAD"});
    pressedEvents.push_back({"place_on_z_numpad", "PLACE_ON_Z_NUMPAD"});

    pressedEvents.push_back({"avi_manipulate_precise", "MANIPULATE_PRECISE"});
    pressedEvents.push_back({"avi_manipulate_very_precise", "MANIPULATE_VERY_PRECISE"});

    pressedEvents.push_back({"avi_camera_mouse", "CAMERA_CONTROL_MOUSE"});
    pressedEvents.push_back({"avi_camera_alt", "CAMERA_CONTROL_ALT"});

    pressedEvents.push_back({"avi_camera_orbit_negative_horizontal_shift", "ORBIT_NEGATIVE_HORIZONTAL_SHIFT"});

    pressedEvents.push_back({"avi_camera_orbit_negative_horizontal_shift", "ORBIT_NEGATIVE_HORIZONTAL_SHIFT"});
    pressedEvents.push_back({"avi_camera_orbit_positive_horizontal_shift", "ORBIT_POSITIVE_HORIZONTAL_SHIFT"});
    pressedEvents.push_back({"avi_camera_orbit_negative_vertical_shift", "ORBIT_NEGATIVE_VERTICAL_SHIFT"});
    pressedEvents.push_back({"avi_camera_orbit_positive_vertical_shift", "ORBIT_POSITIVE_VERTICAL_SHIFT"});

    pressedEvents.push_back({"place_on_x_numpad", "PLACE_ON_X_NUMPAD"});
    pressedEvents.push_back({"place_on_y_numpad", "PLACE_ON_Y_NUMPAD"});
    pressedEvents.push_back({"place_on_z_numpad", "PLACE_ON_Z_NUMPAD"});

    releasedEvents.push_back(
        {"avi_camera_orbit_negative_horizontal", "ORBIT_NEGATIVE_HORIZONTAL_RELEASE"});
    releasedEvents.push_back(
        {"avi_camera_orbit_positive_horizontal", "ORBIT_NEGATIVE_HORIZONTAL_RELEASE"});
    releasedEvents.push_back(
        {"avi_camera_orbit_negative_vertical", "ORBIT_NEGATIVE_HORIZONTAL_RELEASE"});
    releasedEvents.push_back(
        {"avi_camera_orbit_positive_vertical", "ORBIT_NEGATIVE_HORIZONTAL_RELEASE"});

    pressedEvents.push_back({"avi_open_screenshot_directory","OPEN_SCREENSHOT_DIRECTORY"});
    pressedEvents.push_back({"avi_open_video_directory","OPEN_VIDEO_DIRECTORY"});
    pressedEvents.push_back({"avi_open_orientation","ORIENTATION"});
    pressedEvents.push_back({"avi_open_settings","SETTINGS"});
    pressedEvents.push_back({"avi_open_help","HELP"});
}

void
avi::app::KeyInputEventDispatch::hookup(avi::core::EventInvokingSystem& events)
{
    DETOOLIFY;
    this->eventSystem = &events;
}

void
avi::app::KeyInputEventDispatch::_unhandled_input(const godot::Ref<godot::InputEvent>& input)
{
    DETOOLIFY;
    for (auto const& event : pressedEvents)
    {
        if (input->is_action_pressed(event.inputAction))
        {
            eventSystem->invoke(event.eventKey);
        }
    }

    for (auto const& event : releasedEvents)
    {
        if (input->is_action_released(event.inputAction))
        {
            eventSystem->invoke(event.eventKey);
        }
    }
}

std::optional<godot::InputEventKey*>
avi::app::KeyInputEventDispatch::getEventKey(const std::string& eventKey) const
{
    const char* action = nullptr;

    auto findAction = [&eventKey](const std::vector<Event>& events) -> const char*
    {
        for (auto const& event : events)
        {
            if (event.eventKey == eventKey)
            {
                return event.inputAction;
            }
        }

        return nullptr;
    };

    auto setActionFrom = [&action, &findAction](const std::vector<Event>& events)
    {
        if (action == nullptr)
        {
            action = findAction(events);
        }
    };

    setActionFrom(pressedEvents);
    setActionFrom(releasedEvents);

    if (action == nullptr)
    {
        return std::nullopt;
    }

    godot::TypedArray<godot::InputEvent> actionInputList = godot::InputMap::get_singleton()->action_get_events(action);
    for (int index = 0; index < actionInputList.size(); index++)
    {
        godot::Variant value = actionInputList[index];
        auto* inputEvent = godot::Object::cast_to<godot::InputEventKey>(value);

        if (inputEvent != nullptr)
        {
            return inputEvent;
        }
    }

    return std::nullopt;
}
