#pragma once

#include <godot_cpp/classes/node.hpp>
#include <godot_cpp/godot.hpp>

#include <arviz_godot/lib/avi/app/ApplicationSettings.h>
#include <arviz_godot/lib/avi/app/ArVizInteractive.h>
#include <arviz_godot/lib/avi/app/EventList.h>
#include <arviz_godot/lib/avi/app/KeyInputEventDispatch.h>
#include <arviz_godot/lib/avi/app/RevisionLoadCycle.h>
#include <arviz_godot/lib/avi/connection/RevisionBuffer.h>
#include <arviz_godot/lib/avi/interaction/InteractionDispatch.h>
#include <arviz_godot/lib/avi/recording/ImageRecorder.h>
#include <arviz_godot/lib/avi/recording/VideoRecorder.h>
#include <arviz_godot/lib/avi/replay/ReplayController.h>
#include <arviz_godot/lib/avi/scene/CameraSet.h>
#include <arviz_godot/lib/avi/scene/GraphicsController.h>
#include <arviz_godot/lib/avi/scene/ProfileStorage.h>
#include <arviz_godot/lib/avi/scene/ViewCamera.h>
#include <arviz_godot/lib/avi/core/Detoolifier.h>

namespace avi::app
{
    class Main : public godot::Node3D
    {
        GDCLASS(Main, Node3D)

    public:
        Main();

        static void _bind_methods();

        static Main* get();

        void _ready() override;


        void _process(double delta) override;
        void _notification(int what);

    private:
        bool isInitialized = false;
        bool isReady = false;
        bool isExited = false;

        godot::Vector2 minimumWindowSize = godot::Vector2(512, 512);

        avi::app::ArVizInteractive* component{nullptr};

        std::unique_ptr<avi::scene::ProfileStorage> profileStorage{};
        std::unique_ptr<avi::scene::GraphicsController> graphicsController{};
        std::unique_ptr<avi::replay::ReplayController> replayController{};

        avi::scene::ViewCamera* viewCamera;

        std::unique_ptr<avi::app::EventList> eventList{};
        avi::app::KeyInputEventDispatch* keyInputEventDispatch{};

        std::unique_ptr<avi::app::ApplicationSettings> appSettings{};

        std::unique_ptr<avi::connection::RevisionBuffer> revisions{};
        std::unique_ptr<avi::app::RevisionLoadCycle> revisionLoadCycle{};

        std::unique_ptr<avi::scene::CameraSet> cameraSet{};
        avi::interaction::InteractionDispatch* interactionDispatch{};

        std::unique_ptr<avi::recording::ImageRecorder> imageRecorder{};
        std::unique_ptr<avi::recording::VideoRecorder> videoRecorder{};

        static Main* instance;
    };
} // namespace avi::app
