#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "Clock.h"

#include <string>

#include <godot_cpp/classes/os.hpp>

void
avi::connection::Clock::setTimeFrame(avi::connection::TimeFrame frame)
{
    DETOOLIFY;
    fullFrame = frame;
    setAvailableTimeFrame(TimeFrame(0, 0));
    setTargetTime(0);
}

void avi::connection::Clock::setRevisionFrame(long firstRevision, long lastRevision) {
    this->firstRevision = firstRevision;
    this->lastRevision = lastRevision;
}

avi::connection::TimeFrame
avi::connection::Clock::getTimeFrame()
{
    return fullFrame;
}

void
avi::connection::Clock::setAvailableTimeFrame(avi::connection::TimeFrame frame)
{
    DETOOLIFY;
    availableFrame = frame;
}

avi::connection::TimeFrame
avi::connection::Clock::getAvailableTimeFrame()
{
    return availableFrame;
}

void
avi::connection::Clock::advance(long time, bool absolute)
{
    DETOOLIFY;
    float currentSpeed = speed;
    if(absolute) currentSpeed = 1;

    if((fullFrame.begin + targetTime + static_cast<long>(static_cast<long double>(currentSpeed) * time) > fullFrame.end)) {
        targetTime = (fullFrame.end - fullFrame.begin);
    } else if (fullFrame.begin + targetTime + static_cast<long>(static_cast<long double>(currentSpeed) * time) < fullFrame.begin) {
        targetTime = 0;
    } else {
        targetTime += static_cast<long>(static_cast<long double>(currentSpeed) * time);
    }

     checkLoopBack();
}

void avi::connection::Clock::setTargetRevision(long revision) {
    if(revision >= lastRevision || revision < firstRevision) {
        godot::PackedStringArray poolStringArray;
        poolStringArray.push_back("Invalid Revision!");
        std::string notificationString = "Revision " + std::to_string(revision) + " is not in range [" + std::to_string(firstRevision) + "," + std::to_string(lastRevision) + ")";
        poolStringArray.push_back(notificationString.c_str());
        godot::OS::get_singleton()->execute("notify-send", poolStringArray);
        return;
    }

    targetRevision = revision;
    targetRevisionUpdate = true;
}

void avi::connection::Clock::updateTargetRevision(long revision) {
    targetRevision = revision;
}

void
avi::connection::Clock::setTargetTime(long time)
{
    DETOOLIFY;
    if (time <= 0 || (time >= (fullFrame.end - fullFrame.begin) && loopBack)) {
        targetTime = 0;
        clearScene = true;
    } else if (time > (fullFrame.end - fullFrame.begin)) {
        targetTime  = (fullFrame.end - fullFrame.begin);
    } else {
        targetTime = time;
    }

    targetTimeUpdate = true;
}

long
avi::connection::Clock::getTargetTime() const
{
    return targetTime;
}

long avi::connection::Clock::getTargetRevision() const {
    return targetRevision;
}

void
avi::connection::Clock::setSpeed(float newSpeed)
{
    DETOOLIFY;
    speed = newSpeed;
}

float
avi::connection::Clock::getSpeed() const
{
    return speed;
}

void avi::connection::Clock::setLoopBack(bool loopBack) {
    this->loopBack = loopBack;

    checkLoopBack();
}

void avi::connection::Clock::checkLoopBack() {
    if(loopBack && (fullFrame.end - (targetTime + fullFrame.begin)) == 0) {
        setTargetTime(0);
    }
}

size_t avi::connection::Clock::currentlyLoadingBatches() {
    return batchesCurrentlyLoading;
}

void avi::connection::Clock::setLoadingBatches(size_t batchesLoading) {
    batchesCurrentlyLoading = batchesLoading;
}

bool avi::connection::Clock::hasRevisionUpdate() {
    bool update = targetRevisionUpdate;
    targetRevisionUpdate = false;
    return update;
}

bool avi::connection::Clock::hasTargetTimeUpdate() {
    bool update = targetTimeUpdate;
    targetTimeUpdate = false;

    return update;
}

bool avi::connection::Clock::replayClearScene() {
    bool temp = clearScene;
    clearScene = false;

    return temp;
}