#pragma once

#include "TimeFrame.h"

namespace avi::connection
{
    class Clock
    {
    public:
        Clock() = default;

        void setTimeFrame(avi::connection::TimeFrame frame);

        void setRevisionFrame(long firstRevision, long lastRevision);

        avi::connection::TimeFrame getTimeFrame();

        void setAvailableTimeFrame(avi::connection::TimeFrame frame);

        avi::connection::TimeFrame getAvailableTimeFrame();

        [[nodiscard]] float getSpeed() const;

        void advance(long time, bool absolute);

        void setTargetTime(long time);

        void setTargetRevision(long revision);

        bool hasRevisionUpdate();

        bool hasTargetTimeUpdate();

        void updateTargetRevision(long revision);

        [[nodiscard]] long getTargetRevision() const;

        [[nodiscard]] long getTargetTime() const;

        void setSpeed(float newSpeed);

        void setLoopBack(bool loopBack);

        void checkLoopBack();

        size_t currentlyLoadingBatches();

        void setLoadingBatches(size_t batchesLoading);

        bool replayClearScene();

    private:
        avi::connection::TimeFrame fullFrame = avi::connection::TimeFrame(0, 0);
        avi::connection::TimeFrame availableFrame = avi::connection::TimeFrame(0, 0);

        long firstRevision = 0;
        long lastRevision = 0;

        long targetRevision = 0;
        long targetTime = 0;
        float speed = 0.0;
        bool targetRevisionUpdate = false;

        bool targetTimeUpdate = false;

        bool clearScene = false;

        float loopBack = false;

        size_t batchesCurrentlyLoading = 0;
    };
} // namespace avi::connection
