#pragma once

#include <RobotAPI/components/ArViz/ArVizStorage.h>

#include <arviz_godot/lib/avi/connection/Component.h>
#include <arviz_godot/lib/avi/connection/Connectable.h>
#include <arviz_godot/lib/avi/connection/ElementSource.h>
#include <arviz_godot/lib/avi/connection/InteractionSink.h>
#include <arviz_godot/lib/avi/connection/StorageAdapter.h>
#include <arviz_godot/lib/avi/connection/StorageModeSwitch.h>

namespace avi::connection
{
    class ElementStorage :
        public avi::connection::ElementSource,
        public avi::connection::InteractionSink,
        public avi::connection::StorageModeSwitch,
        public avi::connection::Connectable
    {
    public:
        explicit ElementStorage(avi::connection::Component& component);
        ~ElementStorage() override = default;

        avi::connection::StorageAdapter::LayerUpdates getNext() override;

        [[nodiscard]] bool isConnected() const override;

        [[nodiscard]] bool isLive() const override;

        [[nodiscard]] bool isRecording() const override;

        std::string recordReplay(const std::string& name) override;

        void endRecording() override;

        virtual StorageModeSwitch::recordingInfo getRecordingInformation(const std::string name);

        std::vector<std::string> getAllReplays() override;

        std::filesystem::path getRecordingPath() override;

        void useLiveTiming() override;

        void useReplayTiming(const std::string& replay,
                             std::shared_ptr<avi::connection::Clock> clock) override;

        void addInteraction(Interaction interaction) override;

    private:
        avi::connection::Component& component;

        bool live{true};

        bool recording{false};

        bool connected = false;

        std::unique_ptr<avi::connection::StorageAdapter> storage{};

        std::vector<avi::connection::Interaction> feedbackVec{};

        std::mutex feedbackMutex{};

        bool switchToReplay = false;
        bool switchToLive = false;

        std::string replay{};

        std::filesystem::path recordingPath;

        std::shared_ptr<avi::connection::Clock> clock{};
    };
} // namespace avi::connection
