#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "MockStorage.h"

avi::connection::MockStorage::MockStorage() : ElementStorage(this->component)
{
    DETOOLIFY;
}

avi::connection::StorageAdapter::LayerUpdates
avi::connection::MockStorage::getNext()
{
    armarx::viz::data::LayerUpdates layerUpdates;

    armarx::viz::data::LayerUpdate layerUpdate;
    layerUpdate.component = "Component";
    layerUpdate.name = "Layer 1";
    layerUpdate.action = armarx::viz::data::Layer_CREATE_OR_UPDATE;

    IceUtil::Handle<armarx::viz::data::ElementObject> element(
        new armarx::viz::data::ElementObject());
    element->id = "Box";
    element->project = "PriorKnowledgeData";
    element->filename =
        "PriorKnowledgeData/objects/Maintenance/bauhaus-screwbox-large/bauhaus-screwbox-large.xml";
    element->pose = {0, 100, 0, 1, 0, 0, 0};
    setStandardParameter(element, element->id);
    layerUpdate.elements.push_back(element);

    layerUpdates.updates.push_back(layerUpdate);

    return {layerUpdates,false};
}

bool
avi::connection::MockStorage::isConnected() const
{
    return true;
}

bool
avi::connection::MockStorage::isLive() const
{
    return live;
}

std::string
avi::connection::MockStorage::recordReplay(const std::string& name)
{
    std::cout << "MockStorage::recordReplay(" << name << ")" << std::endl;
    return std::string();
}

void
avi::connection::MockStorage::endRecording()
{
    DETOOLIFY;
    std::cout << "MockStorage::endRecording()" << std::endl;
}

std::vector<std::string>
avi::connection::MockStorage::getAllReplays()
{
    std::vector<std::string> result;
    result.emplace_back("replay");
    return result;
}

void
avi::connection::MockStorage::useLiveTiming()
{
    DETOOLIFY;
    live = true;

    std::cout << "MockStorage::useLiveTiming()" << std::endl;
}

void
avi::connection::MockStorage::useReplayTiming(const std::string& replay,
                                              std::shared_ptr<avi::connection::Clock> clock)
{
    DETOOLIFY;
    live = false;

    std::cout << "MockStorage::useReplayTiming(" << replay << ")" << std::endl;
}

void
avi::connection::MockStorage::addInteraction(avi::connection::Interaction interaction)
{
    DETOOLIFY;
    std::cout << "MockStorage::addInteraction(" << interaction.getElement()->getName() << ")"
              << std::endl;
    std::cout << "interaction =" << std::endl;
    std::cout << "type:        " << std::to_string((int)interaction.getType()) << std::endl;
    std::cout << "position:    "
              << "(" << std::to_string(interaction.getTranslation().x) << ", "
              << std::to_string(interaction.getTranslation().y) << ", "
              << std::to_string(interaction.getTranslation().z) << ")" << std::endl;
    std::cout << "transform:   " << std::to_string((int)interaction.getTransformation())
              << std::endl;
}

void
avi::connection::MockStorage::setStandardParameter(
    IceUtil::Handle<armarx::viz::data::Element> element,
    std::basic_string<char> id)
{
    DETOOLIFY;
    element->scale = {1, 1, 1};
    //element->color = {255, 255, 0, 0};
    element->interaction.contextMenuOptions = {id,
                                               "flags:",
                                               "SELECT",
                                               "CONTEXT_MENU",
                                               "TRANSLATION_X",
                                               "implicit global manipulation and transform show"};
    element->interaction.enableFlags = element->interaction.enableFlags |
                                       armarx::viz::data::InteractionEnableFlags::SELECT |
                                       armarx::viz::data::InteractionEnableFlags::CONTEXT_MENU |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_X |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_Y |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_Z |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_X |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_Y |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_Z |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_X |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_Y |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_Z;
}

std::filesystem::path
avi::connection::MockStorage::getRecordingPath()
{
    return ElementStorage::getRecordingPath();
}
bool
avi::connection::MockStorage::isRecording() const
{
    return false;
}

avi::connection::StorageModeSwitch::recordingInfo
avi::connection::MockStorage::getRecordingInformation(const std::string /*name*/)
{
    return {};
}

bool
avi::connection::MockStorage::MockComponent::isConnected() const
{
    return true;
}

armarx::viz::StorageInterfacePrx
avi::connection::MockStorage::MockComponent::getStorageInterface() const
{
    return {};
}
