#pragma once

#include <list>

#include <arviz_godot/lib/avi/connection/Clock.h>
#include <arviz_godot/lib/avi/connection/Component.h>
#include <arviz_godot/lib/avi/connection/StorageAdapter.h>

namespace avi::connection
{
    class ReplayStorageAdapter : public avi::connection::StorageAdapter
    {
    public:
        void setInteraction(std::vector<avi::connection::Interaction> newInteractions) override;

        explicit ReplayStorageAdapter(avi::connection::Component& component,
                                      std::shared_ptr<avi::connection::Clock> clock,
                                      std::string replay);

        ~ReplayStorageAdapter() override = default;

        StorageAdapter::LayerUpdates getLayerUpdates() override;

        std::vector<std::string> getAllReplays() override;

        void setup() override;

        bool finishedProcessing() override;

        void skipToRevision(long revision);

        int64_t getBatchIndex(int64_t timeStamp);

    private:
        StorageAdapter::LayerUpdates getUpdates(int index, int64_t wantedTimeStamp);

        void asyncLoadBatch();

        void updateBatchLoadQueue(int index);

        void onGetBatchAsync(armarx::viz::data::RecordingBatch const& batch);

        void updateBatchCache(int targetIndex);

        struct ArVizWidgetBatchCallback : IceUtil::Shared
        {
            struct ReplayStorageAdapter* this_;

            void
            onSuccess(armarx::viz::data::RecordingBatch const& batch)
            {
                this_->onGetBatchAsync(batch);
            }

            void
            onFailure(Ice::Exception const& ex)
            {
                ARMARX_WARNING << "Failed to get batch async.\nReason:" << ex;
            }
        };


        struct TimestampedRecordingBatch
        {
            IceUtil::Time lastUsed;
            armarx::viz::data::RecordingBatch data;
        };

        struct AsyncBatch
        {
            armarx::viz::data::RecordingBatch batch;
            long batchIndex;
        };

        struct batchInfo {
            int64_t firstTimestamp;
            int64_t lastTimestamp;
            int64_t firstRevision;
            int64_t lastRevision;
            int64_t batchIndex;
        };

        avi::connection::Component& component;

        std::shared_ptr<avi::connection::Clock> clock;

        std::string activeReplay;

        std::map<int64_t, TimestampedRecordingBatch> batchCache{};

        std::vector<AsyncBatch> asyncReceivedBatches{};

        std::vector<long> invalidBatches{};

        std::mutex batchMutex;

        std::vector<batchInfo> batches{};

        long batchCount;

        std::vector<int64_t> batchLoadQueue{};

        bool asyncBatchLoading = false;
        bool discardLoadingBatch = false;

        bool buffering = false;

        int64_t lastSentBatch = 1;

        IceUtil::Handle<ArVizWidgetBatchCallback> callbackData;
        armarx::viz::Callback_StorageInterface_getRecordingBatchPtr callback;
        Ice::AsyncResultPtr callbackResult;

        static size_t
        binarySearch(const std::vector<armarx::viz::data::TimestampedLayerUpdate>& updates,
                     size_t l,
                     size_t r,
                     int64_t timeStamp);

        std::chrono::steady_clock::time_point last = std::chrono::steady_clock::now();

        const static int64_t BATCHES_FRONT = 5;

        const static int64_t BATCHES_BACK = 0;

        const static int64_t BATCH_CACHE_SIZE = BATCHES_FRONT + BATCHES_BACK;
    };
} // namespace avi::connection
