#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "RevisionBuffer.h"

void
avi::connection::RevisionBuffer::addUpdates(avi::connection::StorageAdapter::LayerUpdates&& revision)
{
    DETOOLIFY;
    this->clearScene = revision.clearScene;
    std::vector<armarx::viz::data::LayerUpdate> newUpdates = std::move(revision.updates.updates);
    bufferLock.lock();

    for (armarx::viz::data::LayerUpdate& update : newUpdates)
    {
        LayerRevision layerRevision = {update.component, update.name};

        changed.insert(layerRevision);
        updates[layerRevision] = std::move(update);
    }

    bufferLock.unlock();
}

std::set<avi::connection::LayerRevision>
avi::connection::RevisionBuffer::getChangedLayers()
{
    std::lock_guard<std::mutex> lock(bufferLock);
    return changed;
}

std::optional<armarx::viz::data::LayerUpdate>
avi::connection::RevisionBuffer::getUpdate(const LayerRevision& layer)
{
    std::lock_guard<std::mutex> lock(bufferLock);

    auto found = updates.find(layer);

    if (found != updates.end())
    {
        changed.erase(layer);

        armarx::viz::data::LayerUpdate update = std::move(found->second);
        updates.erase(found);

        return {std::move(update)};
    }

    return std::nullopt;
}

void
avi::connection::RevisionBuffer::clearBuffer()
{
    DETOOLIFY;
    std::lock_guard<std::mutex> lock(bufferLock);
    changed.clear();
    updates.clear();
}

bool avi::connection::RevisionBuffer::getClearScene() {
    return clearScene;
}
