#pragma once

#include <map>
#include <mutex>
#include <optional>
#include <set>

#include <RobotAPI/components/ArViz/ArVizStorage.h>

#include "StorageAdapter.h"

namespace avi::connection
{
    using LayerRevision = std::pair<std::string, std::string>;

    class RevisionBuffer
    {
    public:
        RevisionBuffer() = default;

        void addUpdates(avi::connection::StorageAdapter::LayerUpdates&& revision);

        std::set<LayerRevision> getChangedLayers();

        std::optional<armarx::viz::data::LayerUpdate> getUpdate(const LayerRevision& layer);

        bool getClearScene();

        void clearBuffer();

    private:
        std::map<LayerRevision, armarx::viz::data::LayerUpdate> updates;

        std::set<LayerRevision> changed;

        std::mutex bufferLock;

        bool clearScene = false;
    };
} // namespace avi::connection
