#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ControlAdvertiser.h"
#include "Utilities.h"

#include <functional>
#include <sstream>

size_t
avi::core::ControlAdvertiser::registerAdvertisement(std::vector<Advertisement>&& entries)
{
    size_t index = ads.size();

    std::string text;
    bool isFirst = true;

    for (const auto& entry : entries)
    {
        if (!isFirst)
        {
            text += " | ";
        }

        isFirst = false;

        text += entry.toString(*eventToKeyProvider);
    }

    ads.push_back(text);

    return index;
}

void
avi::core::ControlAdvertiser::activateAdvertisements(size_t index)
{
    DETOOLIFY;
    active.insert(index);
    updateAdvertisements();
}

void
avi::core::ControlAdvertiser::deactivateAdvertisements(size_t index)
{
    DETOOLIFY;
    active.erase(index);
    updateAdvertisements();
}

void
avi::core::ControlAdvertiser::updateAdvertisements()
{
    DETOOLIFY;
    std::vector<std::reference_wrapper<const std::string>> advertisements;

    for (size_t index : active)
    {
        advertisements.emplace_back(ads[index]);
    }

    updateAdvertisements(advertisements);
}

void
avi::core::ControlAdvertiser::setup(avi::core::EventToKeyProvider& eventToKey)
{
    DETOOLIFY;
    this->eventToKeyProvider = &eventToKey;
}

std::string
avi::core::ControlAdvertiser::Advertisement::toString(EventToKeyProvider& eventToKey) const
{
    if (const auto* string = std::get_if<std::string>(&data))
    {
        return *string;
    }

    if (const auto* pair = std::get_if<std::pair<std::string, std::string>>(&data))
    {
        auto const& [name, event] = *pair;

        std::stringstream text;

        auto key = eventToKey.getEventKey(event);
        if (key.has_value())
        {
            text << avi::core::utilities::to_string(key.value()->as_text());
            text << ": ";
        }

        text << name;

        return text.str();
    }

    return "";
}

avi::core::ControlAdvertiser::Advertisement
avi::core::ControlAdvertiser::Advertisement::fromString(std::string string)
{
    Advertisement advertisement;
    advertisement.data = string;

    return advertisement;
}

avi::core::ControlAdvertiser::Advertisement
avi::core::ControlAdvertiser::Advertisement::fromString(std::string action, std::string event)
{
    Advertisement advertisement;
    advertisement.data = action + ": " + event;

    return advertisement;
}


avi::core::ControlAdvertiser::Advertisement
avi::core::ControlAdvertiser::Advertisement::fromEvent(const std::string& name,
                                                       const std::string& event)
{
    Advertisement advertisement;
    advertisement.data = std::make_pair(name, event);

    return advertisement;
}
