#ifndef ARVIZ_GODOT_CONTROLADVERTISER_H
#define ARVIZ_GODOT_CONTROLADVERTISER_H

#include <set>
#include <variant>
#include <vector>

#include <arviz_godot/lib/avi/core/EventToKeyProvider.h>

namespace avi::core
{
    class ControlAdvertiser
    {
    public:
        class Advertisement
        {
        private:
            Advertisement() = default;

        public:
            static Advertisement fromString(std::string string);
            static Advertisement fromString(std::string action, std::string description);
            static Advertisement fromEvent(const std::string& name, const std::string& event);

            std::string toString(EventToKeyProvider& eventToKey) const;

        private:
            std::variant<std::string, std::pair<std::string, std::string>> data;
        };

        virtual ~ControlAdvertiser() = default;

        size_t registerAdvertisement(std::vector<Advertisement>&& entries);

        void activateAdvertisements(size_t index);
        void deactivateAdvertisements(size_t index);

        void setup(EventToKeyProvider& eventToKey);

    protected:
        virtual void
        updateAdvertisements(const std::vector<std::reference_wrapper<const std::string>>& ads) = 0;

    private:
        void updateAdvertisements();

        EventToKeyProvider* eventToKeyProvider;

        std::vector<std::string> ads;
        std::set<size_t> active;
    };
} // namespace avi::core

#endif
