#pragma once

#include <string>

#define ADD_OPTION(TYPE, VALUE, NAME, TOOLTIP) \
    emplace_back(std::make_unique<avi::core::ValueOption<TYPE, T>>(#VALUE, NAME, VALUE, TOOLTIP))

namespace avi::core
{
    template <typename T>
    class Option
    {
    protected:
        Option(std::string key, std::string label, std::string tooltip);

    public:
        virtual void accept(T& transformer) = 0;

        /// Get the key used in storage to refer to this option.
        /// The key must be unique and a valid identifier.
        [[nodiscard]] const std::string& getKey() const;

        /// Get the label referring to this option.
        /// The label must be human-readable.
        [[nodiscard]] const std::string& getLabel() const;

        /// Get a tooltip describing this option.
        [[nodiscard]] const std::string& getToolTip() const;

        virtual ~Option() = default;

    private:
        std::string key;
        std::string label;
        std::string tooltip;
    };
} // namespace avi::core

#include "Option.tpp"
