#pragma once

#include <filesystem>
#include <map>
#include <optional>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/variant/color.hpp>

namespace avi::core
{
    class OptionFile
    {
    public:
        OptionFile() = default;

        const std::string& getName();
        void setName(const std::string& name);

        static std::optional<avi::core::OptionFile> load(const std::filesystem::path& path);
        static avi::core::OptionFile create(const std::string& name);
        void save(const std::filesystem::path& path);

        template <class V>
        std::optional<V>
        getValue(const std::string& key)
        {
            auto value = getValue<int64_t>(key);

            if (value.has_value())
            {
                return static_cast<V>(value.value());
            }

            return std::nullopt;
        }

        template <class V>
        void
        setValue(const std::string& key, V value)
        {
            setValue<int64_t>(key, static_cast<int64_t>(value));
        }

#define SPECIALIZE_TYPE(TYPE)                                                                      \
    template <>                                                                                    \
    std::optional<TYPE> avi::core::OptionFile::getValue<TYPE>(const std::string& key);             \
                                                                                                   \
    template <>                                                                                    \
    void avi::core::OptionFile::setValue<TYPE>(const std::string& key, TYPE value);

    private:
        OptionFile(std::string name, std::map<std::string, std::string> options);

    private:
        std::string name;
        std::map<std::string, std::string> options;
    };
} // namespace avi::core

SPECIALIZE_TYPE(int64_t)
SPECIALIZE_TYPE(double)
SPECIALIZE_TYPE(std::string)
SPECIALIZE_TYPE(bool)
SPECIALIZE_TYPE(std::filesystem::path)
SPECIALIZE_TYPE(godot::Color)