#pragma once

#include <filesystem>
#include <vector>

#include <arviz_godot/lib/avi/core/OptionFileReader.h>
#include <arviz_godot/lib/avi/core/OptionFileWriter.h>

namespace avi::core
{
    struct Settings
    {
        void save();

        static avi::core::Settings getDefaultSettings();

        std::string storageIdentifier{};
        int64_t orbitSensitivityTechnical{};
        int64_t zoomSensitivityTechnical{};
        int64_t rollSensitivityTechnical{};
        int64_t moveSpeedTechnical{};
        int64_t mouseSensitivityPlayful{};
        int64_t moveSpeedPlayful{};
        bool smoothPlayful{};
        int64_t accelerationFactorPlayful{};
        int64_t rollSensitivityPlayful{};
        int64_t revisionCycleFrequency{};
        bool alignCameraToHorizon{};
        std::string defaultProfileName{};
        std::filesystem::path profileDirectory{};
        std::filesystem::path screenshotDirectory{};
        std::filesystem::path videoDirectory{};

        template <typename T>
        auto
        getOptions()
        {
            std::vector<std::unique_ptr<avi::core::Option<T>>> options;

            options.ADD_OPTION(std::string, storageIdentifier, "Storage Identifier", "The identifier of the ArVizStorage component.");
            options.ADD_OPTION(int64_t, orbitSensitivityTechnical, "[Technical] Orbit Sensitivity", "Sensitivity of inputs related to orbit in technical mode.");
            options.ADD_OPTION(int64_t, zoomSensitivityTechnical, "[Technical] Zoom Sensitivity", "Sensitivity of inputs related to zoom in technical mode.");
            options.ADD_OPTION(int64_t, rollSensitivityTechnical, "[Technical] Roll Sensitivity", "Sensitivity of inputs related to roll in technical mode.");
            options.ADD_OPTION(int64_t, moveSpeedTechnical, "[Technical] Move Speed", "Move speed in technical mode.");
            options.ADD_OPTION(int64_t, mouseSensitivityPlayful, "[Playful] Mouse Sensitivity", "Sensitivity of mouse inputs in playful mode.");
            options.ADD_OPTION(int64_t, rollSensitivityPlayful, "[Playful] Roll Sensitivity", "Sensitivity of rolling in playful mode.");
            options.ADD_OPTION(int64_t, moveSpeedPlayful, "[Playful] Move Speed", "Base move speed in playful mode.");
            options.ADD_OPTION(bool, smoothPlayful, "[Playful] Input Smoothing", "Whether to smooth playful input. If no, the acceleration value is ignored.");
            options.ADD_OPTION(int64_t,accelerationFactorPlayful, "[Playful] Acceleration Factor", "How strong the input accelerates (inverse smoothing), from 0 to 10000.");
            options.ADD_OPTION(int64_t, revisionCycleFrequency, "Revision Cycle Frequency", "Frequency [Hz] of requests made to ArVizStorage.");
            options.ADD_OPTION(bool, alignCameraToHorizon, "Align Camera To Horizon", "If true, the camera is not allowed to roll, sticking to a Z-aligned gravity vector. If false, 6D movement is possible.");
            options.ADD_OPTION(std::string, defaultProfileName, "Default Profile Name", "The name of the profile that is loaded on application start.");
            options.ADD_OPTION(std::filesystem::path, profileDirectory, "Profile Directory", "Absolute path to the directory containing all profiles.");
            options.ADD_OPTION(std::filesystem::path, screenshotDirectory, "Screenshot Directory", "Absolute path to a directory in which screenshots will be saved.");
            options.ADD_OPTION(std::filesystem::path, videoDirectory, "Video Directory", "Absolute path to a directory in which videos will be saved.");

            return options;
        }

    private:
        explicit Settings(std::filesystem::path path);

        static std::filesystem::path getPath(const char* xdg, const std::string& directoryInHome);

    private:
        std::filesystem::path path;
        std::optional<avi::core::OptionFile> optionFile{};
    };
} // namespace avi::core
