#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "Utilities.h"

real_t
avi::core::utilities::signed_angle(const godot::Vector3& from,
                                   const godot::Vector3& to,
                                   const godot::Vector3& axis)
{
    godot::Vector3 cross_to = from.cross(to);
    real_t unsigned_angle = godot::Math::atan2(cross_to.length(), from.dot(to));
    real_t sign = cross_to.dot(axis);
    return (sign < 0) ? -unsigned_angle : unsigned_angle;
}

godot::Basis
avi::core::utilities::scaled_local(const godot::Basis& basis, const godot::Vector3& scale)
{
    return basis * godot::Basis(scale.x, 0, 0, 0, scale.y, 0, 0, 0, scale.z);
}

bool
avi::core::utilities::is_equal_approx(const godot::Vector3 a, const godot::Vector3 b)
{
    return godot::Math::is_equal_approx(a.x, b.x) && godot::Math::is_equal_approx(a.y, b.y) &&
           godot::Math::is_equal_approx(a.z, b.z);
}

godot::Vector3
avi::core::utilities::normalized_or_zero(const godot::Vector3& x)
{
    if (is_equal_approx(x, godot::Vector3())) return godot::Vector3();

    return x.normalized();
}

std::string avi::core::utilities::to_string(const godot::String& string)
{
    auto utf8 = string.to_utf8_buffer();

    const auto* ptr = reinterpret_cast<const char*>(utf8.ptr());
    const auto size = static_cast<size_t>(utf8.size());

    return { ptr, size };
}
