#pragma once

#include <functional>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/control.hpp>
#include <godot_cpp/classes/line_edit.hpp>
#include <godot_cpp/classes/option_button.hpp>

#include <arviz_godot/lib/avi/gui/VisualOptionTransformer.h>
#include <arviz_godot/lib/avi/scene/GraphicsController.h>
#include <arviz_godot/lib/avi/scene/ProfileStorage.h>

namespace avi::gui
{
    class GraphicsSettings : public godot::Control
    {
        GDCLASS(GraphicsSettings, Control)

    public:
        static void _bind_methods();

        GraphicsSettings() = default;
        virtual ~GraphicsSettings() = default;

        void _ready() override;

        void setup(avi::scene::GraphicsController& controller, avi::scene::ProfileStorage& storage);

    private:
        void _save();

        void _copy();

        void _select(int64_t idx);

        void _change(godot::Variant v);

        void update();

        void setName(const std::string& name);

        std::string getName();

        void setSavedFlag(bool flag);

    private:
        avi::scene::GraphicsController* controller{};
        avi::scene::ProfileStorage* storage{};

        godot::OptionButton* options{};
        godot::Button* saveButton{};
        godot::LineEdit* nameBox{};

        godot::Control* container{};
        std::vector<std::unique_ptr<avi::core::Option<avi::gui::VisualOptionTransformer>>>
            optionStore{};
        std::vector<std::function<void()>> setters{};
    };
} // namespace avi::gui
