#pragma once

#include <mutex>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/control.hpp>
#include <godot_cpp/classes/button.hpp>
#include <godot_cpp/classes/line_edit.hpp>
#include <godot_cpp/classes/packed_scene.hpp>
#include <godot_cpp/classes/resource_loader.hpp>
#include <godot_cpp/classes/texture.hpp>
#include <godot_cpp/classes/tree.hpp>

#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

namespace avi::gui
{
    class LayerSelection : public godot::Control
    {
        GDCLASS(LayerSelection, Control)

    public:
        static void _bind_methods();

        LayerSelection();
        virtual ~LayerSelection() = default;

        void _ready() override;

        void setup(avi::scene::VisualizedScene& scene);

        void clear();

        void lock();

        void unlock();

    private:
        void update_layers();

        void updateActiveComponents();

        void setVisibility(bool visible, int64_t component, int64_t layer);

        void setVisibilityAll(bool visible);

        void setCollapseAll(bool collapsed);

        void filterLayers();

        void buildTree();

        void setFilterVisibility(bool visible);

        struct LayerTreeItem
        {
            godot::TreeItem* treeItem;
            int componentIndex;
            int layerIndex;
            bool active;
        };

        struct ComponentTreeItem
        {
            godot::TreeItem* treeItem;
            bool previousState;
        };

    private:
        godot::Ref<godot::PackedScene> itemScene;
        godot::Ref<godot::PackedScene> layerScene;

        godot::Tree* tree{};
        godot::TreeItem* treeRootNode{};
        std::vector<ComponentTreeItem> componentTreeItems{};
        std::vector<LayerTreeItem> layerTreeItems{};

        godot::Control* container{nullptr};
        godot::Control* hideAllButton{nullptr};
        godot::Control* showAllButton{nullptr};

        godot::Control* collapseAllButton{nullptr};
        godot::Control* expandAllButton{nullptr};

        godot::LineEdit* filterTextBox{nullptr};
        godot::Control* hideFilteredButton{nullptr};
        godot::Control* showFilteredButton{nullptr};

        std::vector<std::string> components{};
        std::vector<std::vector<std::string>> layers{};
        avi::scene::VisualizedScene* scene{nullptr};

        std::mutex mutex;
    };
} // namespace avi::gui
