#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ManipulationMenu.h"

#include <godot_cpp/classes/input_event.hpp>
#include <arviz_godot/lib/avi/core/Utilities.h>

void
avi::gui::ManipulationMenu::_bind_methods()
{
    DETOOLIFY;
}

avi::gui::ManipulationMenu::ManipulationMenu() = default;

void
avi::gui::ManipulationMenu::setup(avi::interaction::Manipulator& manipulator,
                                  avi::core::EventSubscriptionSystem& eventSubscriptionSystem,
                                  avi::core::EventToKeyProvider& eventToKeyProvider)
{
    DETOOLIFY;
    this->manipulator = &manipulator;

    auto setupButton =
        [this, &eventSubscriptionSystem, &eventToKeyProvider](const char* event, int index)
    {
        eventSubscriptionSystem.subscribe(event, [this, index]() -> void { this->select(index); });

        std::string formattedText = avi::core::utilities::to_string(this->get_item_text(index));
        if (not formattedText.empty())
        {
            formattedText += " ";
        }

        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        if (key.has_value())
        {
            formattedText += "[";
            formattedText += avi::core::utilities::to_string(key.value()->as_text());
            formattedText += "]";
        }

        this->set_item_text(index, formattedText.c_str());
    };

    setupButton("MANIPULATOR_NONE", MANIPULATOR_NONE);
    setupButton("MANIPULATOR_MOVE", MANIPULATOR_MOVE);
    setupButton("MANIPULATOR_ROTATE", MANIPULATOR_ROTATE);
    setupButton("MANIPULATOR_SCALE", MANIPULATOR_SCALE);
}

void
avi::gui::ManipulationMenu::_process(double)
{
    DETOOLIFY;
    setActiveManipulator();
}

void
avi::gui::ManipulationMenu::_ready()
{
    DETOOLIFY;
    this->manipulatorCount = this->get_item_count();
}

int
avi::gui::ManipulationMenu::getSelectedItemID()
{
    for (int i = 0; i < manipulatorCount; i++)
    {
        if (this->is_selected(i))
        {
            return i;
        }
    }

    return -1;
}

void
avi::gui::ManipulationMenu::setActiveManipulator()
{
    DETOOLIFY;
    int activeManipulatorID = getSelectedItemID();

    switch (activeManipulatorID)
    {
        case MANIPULATOR_NONE:
            selectionMode = ManipulationMode::NONE;
            manipulator->setManipulationMode(selectionMode);
            break;
        case MANIPULATOR_MOVE:
            selectionMode = ManipulationMode::TRANSLATE;
            manipulator->setManipulationMode(selectionMode);
            break;
        case MANIPULATOR_ROTATE:
            selectionMode = ManipulationMode::ROTATE;
            manipulator->setManipulationMode(selectionMode);
            break;
        case MANIPULATOR_SCALE:
            selectionMode = ManipulationMode::SCALE;
            manipulator->setManipulationMode(selectionMode);
            break;
        default:
            break;
    }
}
