#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ReplayCockpit.h"

#include <iomanip>
#include <iostream>

#include "UserInterface.h"
#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

#include <ArmarXCore/core/time/DateTime.h>
#include <arviz_godot/lib/avi/core/Utilities.h>


void
avi::gui::ReplayCockpit::_bind_methods()
{
    DETOOLIFY;
    godot::ClassDB::bind_method(godot::D_METHOD("_record"), &ReplayCockpit::_record);

    godot::ClassDB::bind_method(godot::D_METHOD("_enterReplayMode"), &ReplayCockpit::_enterReplayMode);

    godot::ClassDB::bind_method(godot::D_METHOD("_select"), &ReplayCockpit::_select);
    godot::ClassDB::bind_method(godot::D_METHOD("_refresh"), &ReplayCockpit::update_replays);
    godot::ClassDB::bind_method(godot::D_METHOD("_openFolder"), &ReplayCockpit::openFolder);

    godot::ClassDB::bind_method(godot::D_METHOD("_play"), &ReplayCockpit::_play);

    godot::ClassDB::bind_method(godot::D_METHOD("_slide"), &ReplayCockpit::_slide);

    godot::ClassDB::bind_method(godot::D_METHOD("_forward1"), &ReplayCockpit::_forward1);
    godot::ClassDB::bind_method(godot::D_METHOD("_forward2"), &ReplayCockpit::_forward2);
    godot::ClassDB::bind_method(godot::D_METHOD("_rewind1"), &ReplayCockpit::_rewind1);
    godot::ClassDB::bind_method(godot::D_METHOD("_rewind2"), &ReplayCockpit::_rewind2);

    godot::ClassDB::bind_method(godot::D_METHOD("_useSlider"), &ReplayCockpit::_useSlider);
    godot::ClassDB::bind_method(godot::D_METHOD("_unUseSlider"), &ReplayCockpit::_unUseSlider);

    godot::ClassDB::bind_method(godot::D_METHOD("_setPlayBackSpeed", "value"), &ReplayCockpit::_setPlayBackSpeed);

    godot::ClassDB::bind_method(godot::D_METHOD("_setRevision", "value"), &ReplayCockpit::_setRevision);

    godot::ClassDB::bind_method(godot::D_METHOD("clearScene"), &ReplayCockpit::clearScene);

    godot::ClassDB::bind_method(godot::D_METHOD("toggleLoopBack"), &ReplayCockpit::toggleLoopBack);
}

void
avi::gui::ReplayCockpit::_ready()
{
    DETOOLIFY;
    playButton = get_node<godot::Button>("Background/Content/ReplayModeContainer/Buttons/Play");
    playButton->connect("pressed", godot::Callable(this, "_play"));

    recordButton = get_node<godot::Button>("Background/Content/RecordingContainer/Record");
    recordButton->connect("pressed", godot::Callable(this, "_record"));

    modeButton = get_node<godot::Button>("Background/Content/SelectReplayContainer/SelectEnter/ChangeMode");
    modeButton->connect("pressed", godot::Callable(this, "_enterReplayMode"));

    clearSceneButton = get_node<godot::CheckBox>("Background/Content/SelectReplayContainer/Select/ClearScene");
    clearSceneButton->connect("pressed", godot::Callable(this, "clearScene"));

    replayName = get_node<godot::LineEdit>("Background/Content/RecordingContainer/ReplayName");

    slider = get_node<godot::HSlider>("Background/Content/ReplayModeContainer/Slider");
    slider->connect("value_changed", godot::Callable(this, "_slide"));
    slider->connect("mouse_entered", godot::Callable(this, "_useSlider"));
    slider->connect("mouse_exited", godot::Callable(this, "_unUseSlider"));

    replayChooser =
        get_node<godot::OptionButton>("Background/Content/SelectReplayContainer/Chooser");
    replayChooser->connect("item_selected", godot::Callable(this, "_select"));

    refresh = get_node<godot::Button>("Background/Content/SelectReplayContainer/Select/Refresh");
    refresh->connect("pressed", godot::Callable(this, "_refresh"));

    openFolderButton =
        get_node<godot::Button>("Background/Content/SelectReplayContainer/Select/OpenFolder");
    openFolderButton->connect("pressed", godot::Callable(this, "_openFolder"));

    forward1Button =
        get_node<godot::Button>("Background/Content/ReplayModeContainer/Buttons/Forward1");
    forward1Button->connect("pressed", godot::Callable(this, "_forward1"));

    forward2Button =
        get_node<godot::Button>("Background/Content/ReplayModeContainer/Buttons/Forward2");
    forward2Button->connect("pressed", godot::Callable(this, "_forward2"));

    rewind1Button =
        get_node<godot::Button>("Background/Content/ReplayModeContainer/Buttons/Rewind1");
    rewind1Button->connect("pressed", godot::Callable(this, "_rewind1"));

    rewind2Button =
        get_node<godot::Button>("Background/Content/ReplayModeContainer/Buttons/Rewind2");
    rewind2Button->connect("pressed", godot::Callable(this, "_rewind2"));

    currentRevision =
        get_node<godot::SpinBox>("Background/Content/ReplayModeContainer/Progress/Revision");
    currentRevision->connect("value_changed", godot::Callable(this, "_setRevision"));

    playBackSpeed =
        get_node<godot::SpinBox>("Background/Content/ReplayModeContainer/PlayBack/PlayBackSpeed");
    playBackSpeed->set_value(1);
    playBackSpeed->connect("value_changed", godot::Callable(this, "_setPlayBackSpeed"));

    loopBackButton =
            get_node<godot::CheckBox>("Background/Content/ReplayModeContainer/TimeInfo/Loopback");
    loopBackButton->connect("pressed", godot::Callable(this, "toggleLoopBack"));

    replayTimeText = get_node<godot::RichTextLabel>("Background/Content/ReplayModeContainer/TimeInfo/Time");
    replayTimeElapsedText = get_node<godot::RichTextLabel>("Background/Content/ReplayModeContainer/TimeInfo/TimeElapsed");

    recordingContainer = get_node<godot::VBoxContainer>("Background/Content/RecordingContainer");
    replayModeContainer = get_node<godot::VBoxContainer>("Background/Content/ReplayModeContainer");
    recordingInformationContainer =
        get_node<godot::VBoxContainer>("Background/Content/RecordingInformationContainer");
    replayModeContainer->set_visible(false);

    replayControlText = get_node<godot::RichTextLabel>("Background/Content/ReplayModeContainer/ReplayControlText");

    this->playButton->set_disabled(true);
    this->forward1Button->set_disabled(true);
    this->forward2Button->set_disabled(true);
    this->rewind1Button->set_disabled(true);
    this->rewind2Button->set_disabled(true);
}

void
avi::gui::ReplayCockpit::setup(avi::replay::ReplayController& controller,
                               avi::scene::VisualizedScene& scene,
                               avi::gui::UserInterface* ui)
{
    DETOOLIFY;
    this->controller = &controller;
    this->scene = &scene;
    this->ui = ui;
    update_replays();

    if (replayChooser->get_item_count() > 0)
    {
        _select(0);
    }
}

void
avi::gui::ReplayCockpit::_record()
{
    DETOOLIFY;
    if (!controller->getRecordingStatus())
    {
        startRecording();
    }
    else
    {
        endRecording();
    }
}

void
avi::gui::ReplayCockpit::startRecording()
{
    DETOOLIFY;
    std::string recordingName = getName();
    if (recordingName.empty())
    {
        recordingName = "recording";
    }

    replayName->set_text(recordingName.c_str());
    replayName->set_editable(false);

    std::string returnString = controller->startRecording(recordingName);
    if (returnString.empty())
    {
        this->recordButton->set_text("Stop Recording");
    } else if(returnString == "NOT_CONNECTED"){
        godot::PackedStringArray poolStringArray;
        poolStringArray.push_back("Couldn't start recording because there is no connection");
        godot::OS::get_singleton()->execute("notify-send", poolStringArray);
    } else {
        replayName->set_text(returnString.c_str());
        replayName->set_editable(false);

        godot::PackedStringArray poolStringArray;
        poolStringArray.push_back("There is already a recording running with ID:");
        poolStringArray.push_back(returnString.c_str());
        godot::OS::get_singleton()->execute("notify-send", poolStringArray);

        this->recordButton->set_text("Stop Recording");
    }
}

void
avi::gui::ReplayCockpit::endRecording()
{
    DETOOLIFY;
    controller->endRecording();
    replayName->set_editable(true);
    update_replays();
    this->recordButton->set_text("Start Recording");
}

void
avi::gui::ReplayCockpit::_select(int64_t i)
{
    DETOOLIFY;
    this->selectedReplay = avi::core::utilities::to_string( replayChooser->get_item_text(i));
    setupRecordingInformation(selectedReplay);
}

void
avi::gui::ReplayCockpit::update_replays()
{
    DETOOLIFY;
    replayChooser->clear();

    std::vector<std::string> recordingNames = controller->getNames();
    recordingsAvailable = not recordingNames.empty();

    int i = 0;
    bool selectedReplayChanged = true;
    for (auto& name : recordingNames)
    {
        replayChooser->add_item(name.c_str());

        if (selectedReplay == name)
        {
            selectedReplayChanged = false;
            replayChooser->select(i);
        }

        i++;
    }

    if (selectedReplayChanged && recordingsAvailable)
    {
        selectedReplay = avi::core::utilities::to_string(replayChooser->get_item_text(0));
        replayChooser->select(0);
    }

    setupRecordingInformation(selectedReplay);
}

void
avi::gui::ReplayCockpit::openFolder()
{
    DETOOLIFY;
    std::filesystem::path replayPath = controller->getReplayPath();
    if (replayPath.empty())
    {
        replayPath = ".";
    }

    godot::PackedStringArray poolStringArray;
    poolStringArray.push_back("-s");
    poolStringArray.push_back(replayPath.c_str());

    godot::OS::get_singleton()->execute("nautilus", poolStringArray);
}

void
avi::gui::ReplayCockpit::_enterReplayMode()
{
    DETOOLIFY;
    if(clearSceneInReplay) {
        this->scene->clearScene();
    }
    ui->lockLayerSelection();
    ui->clear();
    if (!controller->inReplayMode())
    {
        replayModeActive = true;
        ARMARX_INFO << "Entered Replay Mode" << std::endl;
        _select(replayChooser->get_selected_id());
        this->modeButton->set_text("Exit Replay Mode");
        controller->startReplay(selectedReplay);
        if(clearSceneInReplay) {
            this->scene->clearScene();
        }

        this->recordingContainer->set_visible(false);
        this->replayModeContainer->set_visible(true);
        this->replayChooser->set_disabled(true);
        this->refresh->set_disabled(true);
        this->playButton->set_disabled(false);
        this->forward1Button->set_disabled(false);
        this->forward2Button->set_disabled(false);
        this->rewind1Button->set_disabled(false);
        this->rewind2Button->set_disabled(false);
        this->recordButton->set_disabled(true);

        this->playBackSpeed->set_value(controller->getSpeed());
    }
    else
    {
        replayModeActive = false;
        ARMARX_INFO << "Exited Replay Mode" << std::endl;
        this->modeButton->set_text("Enter Replay Mode");
        this->playButton->set_text("Play");
        this->scene->clearScene();
        controller->endReplay();

        this->slider->set_value(0);
        this->recordingContainer->set_visible(true);
        this->replayModeContainer->set_visible(false);
        this->replayChooser->set_disabled(false);
        this->refresh->set_disabled(false);
        this->playButton->set_disabled(true);
        this->forward1Button->set_disabled(true);
        this->forward2Button->set_disabled(true);
        this->rewind1Button->set_disabled(true);
        this->rewind2Button->set_disabled(true);
        this->recordButton->set_disabled(false);
        this->currentRevision->set_value(0);

        this->playBackSpeed->set_value(controller->getSpeed());
    }
    ui->unlockLayerSelection();
}

void
avi::gui::ReplayCockpit::_forward1()
{
    DETOOLIFY;
    long targetTime = controller->getTargetTime();
    long maxTime = controller->getMaxTime() - controller->getMinTime();
    long newTime = targetTime + 5000000; // 5 sec in microseconds

    if(std::abs(targetTime - maxTime) < 500000) return;

    if(newTime > (maxTime)) {
        newTime = maxTime;
    }

    controller->setTime(newTime);
}

void
avi::gui::ReplayCockpit::_forward2()
{
    DETOOLIFY;
    long targetTime = controller->getTargetTime();
    long maxTime = controller->getMaxTime() - controller->getMinTime();
    long newTime = targetTime + 30000000; // 30 sec in microseconds

    if(std::abs(targetTime - maxTime) < 500000) return;

    if(newTime > (maxTime)) {
        newTime = maxTime;
    }

    controller->setTime(newTime);
}

void
avi::gui::ReplayCockpit::_rewind1()
{
    DETOOLIFY;
    long targetTime = controller->getTargetTime();
    long minTime = 0;
    long newTime = targetTime - 5000000; // 5 sec in microseconds

    if(targetTime == minTime) return;

    if(newTime < minTime) {
        newTime = minTime;
    }

    controller->setTime(newTime);
}

void
avi::gui::ReplayCockpit::_rewind2()
{
    DETOOLIFY;
    long targetTime = controller->getTargetTime();
    long minTime = 0;
    long newTime = targetTime - 30000000; // 30 sec in microseconds

    if(targetTime == minTime) return;

    if(newTime < minTime) {
        newTime = minTime;
    }

    controller->setTime(newTime);
}

void
avi::gui::ReplayCockpit::_play()
{
    DETOOLIFY;
    if (!controller->isPlaying())
    {
        this->playButton->set_text("Pause");
        controller->resumePlayback();
        this->playBackSpeed->set_value(controller->getSpeed());
    }
    else
    {
        this->playButton->set_text("Play");
        controller->pausePlayback();
        this->playBackSpeed->set_value(controller->getSpeed());
    }
}

std::string
avi::gui::ReplayCockpit::getName()
{
    godot::String name = replayName->get_text();
    return avi::core::utilities::to_string(name);
}

void
avi::gui::ReplayCockpit::_slide(float value)
{
    DETOOLIFY;
    if (!actualizeSlider) // not true
    {
        if (controller->inReplayMode())
        {
            double value = slider->get_value();
            value =
                value * static_cast<double>(controller->getMaxTime() - controller->getMinTime());
            controller->setTime(static_cast<long>(value));
        }
        else
        {
            this->slider->set_value(0);
        }
    }
}

void
avi::gui::ReplayCockpit::_process(double delta)
{
    DETOOLIFY;
    if (controller->getRecordingStatus() && !controller->isConnected() && !waitForReconnect)
    {
        this->waitForReconnect = true;
        godot::OS* os = godot::OS::get_singleton();
        godot::PackedStringArray poolStringArray;
        poolStringArray.push_back("Lost connection while recording replay!");
        poolStringArray.push_back("Reconnect to recover replay recording!");
        this->recordButton->set_text("No connection");
        this->recordButton->set_disabled(true);

        os->execute("notify-send", poolStringArray);
    }

    if (waitForReconnect && controller->isConnected())
    {
        this->recordButton->set_disabled(false);
        waitForReconnect = false;
        endRecording();
    }

    if (recordingsAvailable && (!controller->getRecordingStatus()) && !controller->inReplayMode())
    {
        this->modeButton->set_disabled(false);
        this->replayChooser->set_disabled(false);
    }
    else if (!recordingsAvailable || controller->getRecordingStatus())
    {
        this->modeButton->set_disabled(true);
        this->replayChooser->set_disabled(true);
    }

    if (controller->inReplayMode() && !sliderUsed)
    {
        actualizeSlider = true;
        slider->set_value(static_cast<float>(
            static_cast<long double>(controller->getTargetTime()) /
            static_cast<long double>(controller->getMaxTime() - controller->getMinTime())));
        actualizeSlider = false;
    }

    if (controller->inReplayMode())
    {
        this->currentRevision->set_value((static_cast<float>(controller->getTargetRevision())));
        updateReplayTimeText();
    }

    if(replayModeActive) {
        size_t loadingBatches = controller->currentlyLoadingBatches();
        if(loadingBatches != 0){
            std::string replayText = "Replay Control | Loading new batches...";
            replayControlText->set_text(replayText.c_str());
        } else {
            replayControlText->set_text("Replay Control");
        }
    }

    currentRevision->set_editable(!controller->isPlaying());
}

void
avi::gui::ReplayCockpit::_useSlider()
{
    DETOOLIFY;
    sliderUsed = true;
    playingBeforeSlider = controller->isPlaying();
    if (controller->inReplayMode())
    {
        controller->pausePlayback();
    }
}

void
avi::gui::ReplayCockpit::_unUseSlider()
{
    DETOOLIFY;
    if (playingBeforeSlider)
    {
        controller->resumePlayback();
    }
    sliderUsed = false;
}

void
avi::gui::ReplayCockpit::_setPlayBackSpeed(godot::Variant)
{
    DETOOLIFY;
    if (controller->inReplayMode())
    {
        controller->setSpeed(playBackSpeed->get_value());
    }
    else
    {
        playBackSpeed->set_value(1);
    }
}

void
avi::gui::ReplayCockpit::_setRevision(godot::Variant)
{
    DETOOLIFY;
    if (controller->inReplayMode() && !controller->isPlaying())
    {
        controller->setRevision((long)(currentRevision->get_value()));
    }
}

void
avi::gui::ReplayCockpit::setupRecordingInformation(const std::string name)
{
    DETOOLIFY;
    connection::StorageModeSwitch::recordingInfo recordingInfo;
    recordingInfo = controller->getRecordingInformation(name);
    firstTimestamp = recordingInfo.firstTimeStamp;
    int informationContainerCount = 8;

    if (recordingInfo.recordingID != "")
    {
        armarx::DateTime firstTimeStamp { armarx::Duration::MicroSeconds(recordingInfo.firstTimeStamp) };
        armarx::DateTime lastTimeStamp { armarx::Duration::MicroSeconds(recordingInfo.lastTimeStamp) };

        for (int containerIndex = 1; containerIndex <= informationContainerCount; containerIndex++)
        {
            godot::Node* node = recordingInformationContainer->get_child(containerIndex);
            godot::RichTextLabel* variable = static_cast<godot::RichTextLabel*>(node->get_child(0));
            godot::RichTextLabel* information =
                static_cast<godot::RichTextLabel*>(node->get_child(1));
            variable->set_text(node->get_name());

            switch (containerIndex)
            {
                case 1:
                    information->set_text(recordingInfo.recordingID.c_str());
                    break;
                case 2:
                    information->set_text(firstTimeStamp.toDateTimeString().c_str());
                    break;
                case 3:
                    information->set_text(lastTimeStamp.toDateTimeString().c_str());
                    break;
                case 4:
                    information->set_text(
                        (lastTimeStamp - firstTimeStamp).toDurationString().c_str());
                    break;
                case 5:
                    information->set_text(std::to_string(recordingInfo.revisionStart).c_str());
                    break;
                case 6:
                    information->set_text(std::to_string(recordingInfo.revisionEnd).c_str());
                    break;
                case 7:
                    information->set_text(
                        std::to_string(recordingInfo.revisionEnd - recordingInfo.revisionStart)
                            .c_str());
                    break;
                case 8:
                    information->set_text(std::to_string(recordingInfo.batches).c_str());
                    break;
            }
        }
    }
    else
    {
        for (int containerIndex = 0; containerIndex < informationContainerCount; containerIndex++)
        {
            godot::Node* node = recordingInformationContainer->get_child(containerIndex);
            if (node->get_child_count() > 1)
            {
                godot::RichTextLabel* information =
                    cast_to<godot::RichTextLabel>(node->get_child(1));
                information->clear();
            }
        }
    }
}

void avi::gui::ReplayCockpit::clearScene() {
    clearSceneInReplay = !clearSceneInReplay;
}

void avi::gui::ReplayCockpit::toggleLoopBack() {
    loopBack = !loopBack;
    controller->setLoopBack(loopBack);
}

void avi::gui::ReplayCockpit::updateReplayTimeText() {
    long targetTime = controller->getTargetTime();

    armarx::DateTime dateTime { armarx::Duration::MicroSeconds(firstTimestamp + targetTime) };
    godot::String timeString = dateTime.toDateTimeString().c_str();
    timeString = timeString.substr(0,23);

    armarx::DateTime dateTimeElapsed {armarx::Duration::MicroSeconds(targetTime)};
    godot::String timeStringElapsed = dateTimeElapsed.toDurationSinceEpoch().toDurationString().c_str();

    godot::String textElapsed = "\n" + timeStringElapsed;
    replayTimeElapsedText->set_text(textElapsed);

    godot::String text = "\n" + timeString;
    replayTimeText->set_text(text);
}






