#pragma once

#include <godot_cpp/classes/check_box.hpp>
#include <godot_cpp/classes/control.hpp>
#include <godot_cpp/classes/h_slider.hpp>
#include <godot_cpp/classes/label.hpp>
#include <godot_cpp/classes/line_edit.hpp>
#include <godot_cpp/classes/rich_text_label.hpp>
#include <godot_cpp/classes/spin_box.hpp>
#include <godot_cpp/classes/v_box_container.hpp>
#include <godot_cpp/godot.hpp>

#include <arviz_godot/lib/avi/replay/ReplayController.h>
#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

namespace avi::gui
{
    class UserInterface;

    class ReplayCockpit : public godot::Control
    {
        GDCLASS(ReplayCockpit, Control)

    public:
        static void _bind_methods();

        ReplayCockpit() = default;
        virtual ~ReplayCockpit() = default;

        void _ready() override;

        void setup(avi::replay::ReplayController& controller,
                   avi::scene::VisualizedScene& scene,
                   avi::gui::UserInterface* ui);

        void _process(double delta) override;

        void _record();

        void _enterReplayMode();

        void _select(int64_t i);

        void _forward1();

        void _forward2();

        void _rewind1();

        void _rewind2();

        void _play();

        void _slide(float value);

        void _useSlider();

        void _unUseSlider();

        void _setPlayBackSpeed(godot::Variant);
        void _setRevision(godot::Variant);

    private:
        void startRecording();

        void endRecording();

        void update_replays();

        void openFolder();

        void clearScene();

        void toggleLoopBack();

        void setupRecordingInformation(const std::string name);

        void updateReplayTimeText();

        std::string getName();

        avi::replay::ReplayController* controller{};
        avi::scene::VisualizedScene* scene{};
        avi::gui::UserInterface* ui{};

        bool sliderUsed = false;
        bool playingBeforeSlider = false;
        bool actualizeSlider = false;
        bool recordingsAvailable = false;

        bool waitForReconnect = false;

        bool clearSceneInReplay = true;
        bool replayModeActive = false;

        bool loopBack = false;
        long firstTimestamp = 0;

        std::string selectedReplay{};

        godot::OptionButton* replayChooser{};
        godot::Button* refresh{};
        godot::Button* openFolderButton{};

        godot::Button* recordButton{};

        godot::Button* modeButton;
        godot::CheckBox* clearSceneButton{};

        godot::Button* playButton{};
        godot::Button* forward1Button{};
        godot::Button* forward2Button{};
        godot::Button* rewind1Button{};
        godot::Button* rewind2Button{};

        godot::CheckBox* loopBackButton{};
        godot::RichTextLabel* replayTimeElapsedText{};
        godot::RichTextLabel* replayTimeText{};

        godot::LineEdit* replayName{};

        godot::HSlider* slider{};

        godot::SpinBox* currentRevision{};

        godot::SpinBox* playBackSpeed{};

        godot::VBoxContainer* recordingContainer{};
        godot::VBoxContainer* replayModeContainer{};
        godot::VBoxContainer* recordingInformationContainer{};

        godot::RichTextLabel* replayControlText{};
    };
} // namespace avi::gui
