#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "SettingsWindow.h"

void
avi::gui::SettingsWindow::_bind_methods()
{
    DETOOLIFY;
    godot::ClassDB::bind_method(godot::D_METHOD("_change"), &SettingsWindow::_change);
    godot::ClassDB::bind_method(godot::D_METHOD("_save"), &SettingsWindow::_save);
}

void
avi::gui::SettingsWindow::setup(avi::core::Settings& settings)
{
    DETOOLIFY;
    this->settings = &settings;

    saveButton = get_node<godot::Button>("Container/Save");
    saveButton->connect("pressed", godot::Callable(this, "_save"));

    options = settings.getOptions<VisualOptionTransformer>();
    std::vector<avi::core::Option<VisualOptionTransformer>*> toTransform;
    toTransform.reserve(options.size());
    for (auto& option : options)
    {
        toTransform.push_back(option.get());
    }

    VisualOptionTransformer transformer;
    setters = transformer.createGraphicsControls(
        get_node<godot::Control>("Container/Scroll/Content"), this, std::move(toTransform));
}

void
avi::gui::SettingsWindow::open()
{
    DETOOLIFY;
    setSavedFlag(true);
    this->popup_centered();
}

void
avi::gui::SettingsWindow::_change(godot::Variant v)
{
    DETOOLIFY;
    for (auto& setter : setters)
    {
        setter();
    }

    setSavedFlag(false);
}

void
avi::gui::SettingsWindow::_save()
{
    DETOOLIFY;
    setSavedFlag(true);

    settings->save();
    this->hide();
}

void
avi::gui::SettingsWindow::setSavedFlag(bool flag)
{
    DETOOLIFY;
    saveButton->set_text(flag ? "Save" : "Save (*)");
}
