#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "StatusIcons.h"

#include <godot_cpp/classes/input_event_key.hpp>

void
avi::gui::StatusIcons::_bind_methods()
{
    DETOOLIFY;

}

avi::gui::StatusIcons::StatusIcons() = default;

void
avi::gui::StatusIcons::setup(avi::connection::Connectable& source,
                             avi::connection::StorageModeSwitch& mode,
                             avi::recording::VideoRecorder& recorder,
                             avi::scene::GraphicsController& graphicsController,
                             avi::core::EventToKeyProvider& eventToKeyProvider)
{
    DETOOLIFY;
    this->source = &source;
    this->mode = &mode;
    this->recorder = &recorder;
    this->graphicsController = &graphicsController;
    this->setup(eventToKeyProvider);

    this->itemList = get_node<godot::ItemList>("Container/List");
    this->disconnectedTexture = this->itemList->get_item_icon(0);
    this->connectedTexture = this->itemList->get_item_icon(1);
    this->replayTexture = this->itemList->get_item_icon(2);
    this->recordingTexture = this->itemList->get_item_icon(3);
    this->replayRecordingTexture = this->itemList->get_item_icon(4);

    this->itemList->clear();
    this->itemList->add_item("Disconnected", this->disconnectedTexture, false);
    this->itemList->add_item(advertisement.c_str());

    this->graphicsController->registerChangeListener(this, [this]() { this->updateTextColor(); });
}

void
avi::gui::StatusIcons::_process(double /*delta*/)
{
    DETOOLIFY;
    if (this->source == nullptr)
    {
        return;
    }

    connectionMode newMode;

    if (mode->isLive())
    {
        if (source->isConnected())
        {
            newMode = CONNECTED;
        }
        else
        {
            newMode = DISCONNECTED;
        }
    }
    else
    {
        newMode = REPLAY;
    }

    bool isRecording = recorder->isRecording();
    bool isRecordingReplay = mode->isRecording();

    if (newMode != this->currentMode || isRecording != this->isVideoRecording ||
        isRecordingReplay != this->isReplayRecording)
    {
        this->currentMode = newMode;
        this->isVideoRecording = isRecording;
        this->isReplayRecording = isRecordingReplay;

        itemList->clear();

        switch (currentMode)
        {
            case DISCONNECTED:
                itemList->add_item("Disconnected", this->disconnectedTexture, false);
                break;
            case CONNECTED:
                itemList->add_item("", this->connectedTexture, false);
                break;
            case REPLAY:
                itemList->add_item("Replay", this->replayTexture, false);
                break;
        }

        if (isVideoRecording)
        {
            itemList->add_item("Recording [V to End]", this->recordingTexture, false);
        }

        if (isReplayRecording)
        {
            itemList->add_item("Recording Replay", this->replayRecordingTexture, false);
        }

        itemList->add_item(advertisement.c_str());

        updateTextColor();
    }
}

void
avi::gui::StatusIcons::updateTextColor()
{
    DETOOLIFY;
    float g = this->graphicsController->getBackgroundColor().inverted().get_v();
    godot::Color newColor(g, g, g);

    for (int item = 0; item < this->itemList->get_item_count(); item++)
    {
        this->itemList->set_item_custom_fg_color(item, newColor);
    }
}

void
avi::gui::StatusIcons::updateAdvertisements(
    const std::vector<std::reference_wrapper<const std::string>>& ads)
{
    DETOOLIFY;

    advertisement = "|| F1: Help || Actions: ";
    bool isFirst = true;

    for (const auto& ad : ads)
    {
        if (!isFirst)
        {
            advertisement += " | ";
        }

        isFirst = false;

        advertisement += ad;
    }

    itemList->set_item_text(itemList->get_item_count() - 1, advertisement.c_str());
}
