#pragma once

#include <godot_cpp/classes/control.hpp>
#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/item_list.hpp>
#include <godot_cpp/classes/label.hpp>
#include <godot_cpp/classes/texture.hpp>

#include <arviz_godot/lib/avi/connection/Connectable.h>
#include <arviz_godot/lib/avi/connection/StorageModeSwitch.h>
#include <arviz_godot/lib/avi/core/ControlAdvertiser.h>
#include <arviz_godot/lib/avi/recording/VideoRecorder.h>

namespace avi::gui
{
    class StatusIcons : public godot::Control, public core::ControlAdvertiser
    {
        GDCLASS(StatusIcons, Control)

    public:
        static void _bind_methods();

        StatusIcons();
        virtual ~StatusIcons() = default;

        void _process(double delta) override;

        using avi::core::ControlAdvertiser::setup;

        void setup(avi::connection::Connectable& source,
                   avi::connection::StorageModeSwitch& mode,
                   avi::recording::VideoRecorder& recorder,
                   avi::scene::GraphicsController& graphicsController,
                   avi::core::EventToKeyProvider& eventToKeyProvider);

    protected:
        void updateAdvertisements(
            const std::vector<std::reference_wrapper<const std::string>>& ads) override;

    private:
        void updateTextColor();

        avi::connection::Connectable* source{};
        avi::connection::StorageModeSwitch* mode{};
        avi::recording::VideoRecorder* recorder{};
        avi::scene::GraphicsController* graphicsController{};

        enum connectionMode
        {
            DISCONNECTED,
            CONNECTED,
            REPLAY,
        };

        godot::Ref<godot::Texture> disconnectedTexture;
        godot::Ref<godot::Texture> connectedTexture;
        godot::Ref<godot::Texture> replayTexture;
        godot::Ref<godot::Texture> recordingTexture;
        godot::Ref<godot::Texture> replayRecordingTexture;

        connectionMode currentMode{};
        bool isVideoRecording{};
        bool isReplayRecording{};

        godot::ItemList* itemList{};
        std::string advertisement{};
    };
} // namespace avi::gui
