#pragma once

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/node.hpp>

#include <arviz_godot/lib/avi/connection/InteractionSink.h>
#include <arviz_godot/lib/avi/connection/StorageModeSwitch.h>
#include <arviz_godot/lib/avi/core/InterfaceUtility.h>
#include <arviz_godot/lib/avi/interaction/Manipulator.h>
#include <arviz_godot/lib/avi/scene/CameraSet.h>
#include <arviz_godot/lib/avi/scene/ElementNode.h>


namespace avi::interaction
{

    class InteractionDispatch : public godot::Node
    {
        GDCLASS(InteractionDispatch, Node)

        void quit();

    public:
        InteractionDispatch();
        virtual ~InteractionDispatch() = default;

        static void _bind_methods();

        void _process(double delta) override;
        void _notification(int what);

        void setupControlAdvertisement(avi::core::ControlAdvertiser& advertiser);

        void _unhandled_input(godot::Ref<godot::InputEvent> const& event) override;

        void setup(avi::connection::InteractionSink& interactionSink,
                   avi::scene::CameraSet& cameraSet,
                   avi::core::InterfaceUtility& ui,
                   avi::connection::StorageModeSwitch& storageMode);

        void sendInteraction(avi::connection::Interaction& interaction);

        void sendManipulation(avi::connection::Interaction& interaction);

        avi::interaction::Manipulator* getManipulator();

    private:
        avi::connection::InteractionSink* interactionSink{};
        avi::connection::StorageModeSwitch* storageMode{};
        avi::scene::CameraSet* cameraSet{};
        avi::core::InterfaceUtility* userInterface{};
        avi::core::Camera* activeCamera;

        std::optional<std::pair<avi::core::Element*, godot::Node*>> targetElement;
        avi::core::Element* previousElement = nullptr;
        avi::connection::Interaction interaction;
        bool wasManipulation = false;

        godot::Control* rootControl;

        void deselect();

        std::unique_ptr<avi::interaction::Manipulator> manipulator{};

        void startInteraction();

        void startManipulation();

        bool enableManipulation = true;
    };
} // namespace avi::interaction
