/*
* Adapted from: godot/editor/plugins/node_3d_editor_plugin.cpp
*/

#include "Interaction.h"

#include <godot_cpp/classes/camera3d.hpp>
#include <godot_cpp/classes/input.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/viewport.hpp>

godot::Interaction
godot::Interaction::compute(const godot::InputEventMouseButton& mouseButtonEvent,
                            const godot::Node3D& node,
                            const Node3D& proxy,
                            const godot::Camera3D& camera,
                            int axis)
{
    godot::Interaction interaction;

    interaction.original_global = node.get_global_transform();
    interaction.original_local = node.get_transform();
    interaction.center_global = node.get_global_transform().origin;

    interaction.proxy_original_global = proxy.get_global_transform();
    interaction.proxy_original_local = proxy.get_transform();

    godot::Vector2 mouse_pos = node.get_viewport()->get_mouse_position();

    interaction.click_ray_normal_global = get_ray_normal(camera, mouse_pos).normalized();
    interaction.click_ray_origin_global = get_ray_origin(camera, mouse_pos);

    interaction.plane = static_cast<TransformPlane>(TRANSFORM_X_AXIS + axis);

    return interaction;
}

godot::Vector3
godot::Interaction::get_ray_origin(const godot::Camera3D& camera, const godot::Vector2& p_pos)
{
    return camera.project_ray_origin(p_pos);
}

godot::Vector3
godot::Interaction::get_ray_normal(const godot::Camera3D& camera, godot::Vector2& p_pos)
{
    return camera.project_ray_normal(p_pos);
}



std::optional<float>
getPrecision(float precise, float morePrecise, float mostPrecise)
{
    godot::Input* input = godot::Input::get_singleton();

    bool precisePressed = input->is_action_pressed("avi_manipulate_precise");
    bool veryPrecisePressed = input->is_action_pressed("avi_manipulate_very_precise");

    if (precisePressed && veryPrecisePressed)
    {
        return morePrecise;
    }

    if (precisePressed)
    {
        return precise;
    }

    if (veryPrecisePressed)
    {
        return mostPrecise;
    }

    return {};
}