#pragma once

#include <optional>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/variant/transform3d.hpp>
#include <godot_cpp/variant/vector2.hpp>
#include <godot_cpp/variant/vector3.hpp>

namespace godot
{
    class Camera3D;
    class InputEventMouseButton;
    class Node3D;

    enum TransformPlane
    {
        TRANSFORM_X_AXIS,
        TRANSFORM_Y_AXIS,
        TRANSFORM_Z_AXIS
    };

    struct Interaction
    {
        TransformPlane plane;

        godot::Transform3D original_global;
        godot::Transform3D original_local;
        Vector3 center_global;

        Vector3 click_ray_normal_global;
        Vector3 click_ray_origin_global;

        godot::Transform3D proxy_original_global;
        godot::Transform3D proxy_original_local;

        static Interaction compute(const InputEventMouseButton& mouseButtonEvent,
                                   const Node3D& node,
                                   const Node3D& proxy,
                                   const Camera3D& camera,
                                   int axis);

        static Vector3 get_ray_origin(const Camera3D& camera, const Vector2& p_pos);
        static Vector3 get_ray_normal(const Camera3D& camera, Vector2& p_pos);
    };

} // namespace godot

std::optional<float> getPrecision(float precise, float more_precise, float most_precise);
