/*
 * Code taken from godot/editor/plugins/spatial_editor_plugin.cpp
 * https://github.com/godotengine/godot/blob/3.4.3-stable/editor/plugins/spatial_editor_plugin.cpp
 *
 * Adapted by Rainer Kartmann
 */

#pragma once

#include <godot_cpp/classes/standard_material3d.hpp>

#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/ref.hpp>
#include <godot_cpp/classes/shader_material.hpp>



namespace godot
{
    class ArrayMesh;
    class ShaderMaterial;
    class StandardMaterial3D;

    struct Manipulator
    {
        struct Parameters
        {
            float alpha = 255;

            Color colors[3] = {
                Color(.965, .282, .392),
                Color(.502, .816, .106),
                Color(.247, .596, .965),
            };

            float gizmo_arrow_size = 0.35;
            float gizmo_ring_half_width = 0.1;
            float gizmo_scale_default = 0.15;
            float gizmo_plane_size = 0.2;
            float gizmo_plane_dst = 0.3;
            float gizmo_circle_size = 1.1;
            float gizmo_scale_offset = (gizmo_circle_size + 0.3);
            float gizmo_arrow_offset = (gizmo_circle_size + 0.3);
        };
        Parameters params;

        Ref<ArrayMesh> move_gizmo[3], move_plane_gizmo[3], rotate_gizmo[4], scale_gizmo[3],
            scale_plane_gizmo[3];

        Ref<StandardMaterial3D> gizmo_color[3];
        Ref<StandardMaterial3D> plane_gizmo_color[3];
        Ref<ShaderMaterial> rotate_gizmo_color[3];
        Ref<StandardMaterial3D> gizmo_color_hl[3];
        Ref<StandardMaterial3D> plane_gizmo_color_hl[3];
        Ref<ShaderMaterial> rotate_gizmo_color_hl[3];

        struct AxisData
        {
            int axis = 0;

            Color col;
            Color albedo;
            Ref<StandardMaterial3D> mat;

            Vector3 ivec;
            Vector3 nivec;
            Vector3 ivec2;
            Vector3 ivec3;
        };
        AxisData make_axis_data(int axis);

        void init_all_gizmos();
        void init_all_gizmos(int axis);

        void init_move_gizmo(const AxisData& d);
        void init_move_plane_gizmo(const AxisData& d);

        void init_rotate_gizmo(const AxisData& d);

        void init_scale_gizmo(const AxisData& d);
        void init_scale_plane_gizmo(const AxisData& d);
    };

} // namespace godot
