#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ArrowCircleElementNode.h"

#include <cmath>

#include <godot_cpp/classes/box_mesh.hpp>
#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::ArrowCircleElementNode::_bind_methods()
{
    DETOOLIFY;

}

avi::scene::ArrowCircleElementNode::ArrowCircleElementNode()
{
    DETOOLIFY;
}

godot::AABB
avi::scene::ArrowCircleElementNode::getVolume()
{
    godot::AABB correctedAABB = torusMeshInstance->get_aabb();
    correctedAABB.position = staticBody->get_transform().origin;

    godot::Vector3 size = torusMeshInstance->get_aabb().get_size() * staticBody->get_scale();
    correctedAABB.size = {size.x * radius * 0.005f, size.y * radius * 0.005f, size.z * 0.5f};

    return correctedAABB;
}

godot::Transform3D
avi::scene::ArrowCircleElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

void
avi::scene::ArrowCircleElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::ArrowCircleElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::ArrowCircleElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::ArrowCircleElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::ArrowCircleElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    auto* mat = cast_to<godot::StandardMaterial3D>(torusMeshInstance->get_material_override().ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}

void
avi::scene::ArrowCircleElementNode::setCompletion(float completion)
{
    DETOOLIFY;
    this->completion = completion;
}

void
avi::scene::ArrowCircleElementNode::setWidth(float width)
{
    DETOOLIFY;
    this->width = width;
}

void
avi::scene::ArrowCircleElementNode::setRadius(float radius)
{
    DETOOLIFY;
    this->radius = radius;
    this->collisionShape->set_scale({radius, 10, radius});
}

godot::Vector3
avi::scene::ArrowCircleElementNode::getPosition()
{
    return torusMeshInstance->get_position();
}

godot::Vector3
avi::scene::ArrowCircleElementNode::getRotation()
{
    return torusMeshInstance->get_rotation();
}

godot::Vector3
avi::scene::ArrowCircleElementNode::getScale()
{
    return torusMeshInstance->get_scale();
}

avi::scene::ArrowCircleElementNode*
avi::scene::ArrowCircleElementNode::newArrowCircleElementNode(avi::connection::SentElementType type,
                                                              std::string name,
                                                              avi::core::Layer* layer)
{
    ArrowCircleElementNode* node = newArrowCircleElementNode();
    node->setAllAttributes(type, name, layer);

    return node;
}

avi::scene::ArrowCircleElementNode*
avi::scene::ArrowCircleElementNode::newArrowCircleElementNode()
{
    ArrowCircleElementNode* node = memnew(ArrowCircleElementNode);
    node->torusMeshInstance = memnew(godot::MeshInstance3D);
    node->staticBody = memnew(godot::StaticBody3D);
    node->collisionShape = memnew(godot::CollisionShape3D);
    godot::StandardMaterial3D* material = memnew(godot::StandardMaterial3D);
    node->cone = memnew(godot::MeshInstance3D);
    godot::CylinderMesh* coneMesh = memnew(godot::CylinderMesh);

    node->cylinderMesh = coneMesh;
    material->set_cull_mode(godot::StandardMaterial3D::CULL_DISABLED);
    node->torusMeshInstance->set_material_override(material);
    node->add_child(node->staticBody);
    node->staticBody->add_child(node->collisionShape);
    node->staticBody->add_child(node->torusMeshInstance);

    coneMesh->set_top_radius(0);
    coneMesh->set_bottom_radius(1);
    coneMesh->set_height(1);

    node->cone->set_mesh(coneMesh);
    node->cone->set_material_override(material);
    node->staticBody->add_child(node->cone);

    return node;
}

void
avi::scene::ArrowCircleElementNode::setAllAttributes(avi::connection::SentElementType type,
                                                     std::string name,
                                                     avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::ArrowCircleElementNode::updateAttributes(int flag,
                                                     std::vector<std::string> interactions,
                                                     godot::Vector3 position,
                                                     godot::Vector3 scale,
                                                     godot::Quaternion rotation,
                                                     godot::Color color,
                                                     float completion,
                                                     float radius,
                                                     float width)
{
    setInteractionFlag(flag);
    if (!shapeChecked)
    {
        if (flag & armarx::viz::data::InteractionEnableFlags::SELECT)
        {
            setShape();
        }
        shapeChecked = true;
    }
    setInteractions(interactions);
    setPosition(position);
    setColor(color);

    float oldCompletion = this->completion;
    float oldRadius = this->radius;
    float oldWidth = this->width;

    setCompletion(completion);
    setWidth(width);
    setRadius(radius);

    if ((completion != oldCompletion) || (radius != oldRadius) || (width != oldWidth))
    {
        generateMesh();
    }

    setRotation(rotation);
    setScale(scale);
}

void
avi::scene::ArrowCircleElementNode::generateMesh()
{
    DETOOLIFY;
    godot::PackedVector3Array vertexPositions;
    godot::PackedVector3Array normals;
    godot::PackedInt32Array faces;

    float completion2 = std::min<float>(1.0f, std::max(-1.0f, completion));
    int sign = completion2 >= 0 ? 1 : -1;
    float torusCompletion = completion2 - 1.0f / RINGS * sign;
    if (torusCompletion * sign < 0)
    {
        torusCompletion = 0;
    }

    const float TWO_PI = 2.0f * (float)M_PI;

    {
        // Create a torus mesh (for the completion circle)
        float tubeRadius = width;

        int numVerticesPerRow = SIDES + 1;
        int numVerticesPerColumn = RINGS + 1;

        int storageSize = numVerticesPerRow * numVerticesPerColumn;
        vertexPositions.resize(storageSize);
        godot::Vector3* positionsPtr = vertexPositions.ptrw();
        normals.resize(storageSize);
        godot::Vector3* normalsPtr = normals.ptrw();

        float theta = 0.0f;
        float phi = 0.0f;
        float verticalAngularStride = TWO_PI / RINGS;
        float horizontalAngularStride = TWO_PI / SIDES;

        size_t writeIndex = 0;

        for (int verticalIt = 0; verticalIt < numVerticesPerColumn; verticalIt++)
        {
            theta = verticalAngularStride * verticalIt * torusCompletion;

            for (int horizontalIt = 0; horizontalIt < numVerticesPerRow; horizontalIt++)
            {
                phi = horizontalAngularStride * horizontalIt;

                // position
                float x = std::cos(theta) * (radius + tubeRadius * std::cos(phi));
                float y = std::sin(theta) * (radius + tubeRadius * std::cos(phi));
                float z = tubeRadius * std::sin(phi);

                float xCenter = std::cos(theta) * radius;
                float yCenter = std::sin(theta) * radius;
                float zCenter = 0;

                godot::Vector3 normal = godot::Vector3(x - xCenter, y - yCenter, z - zCenter);
                normal.normalize();

                positionsPtr[writeIndex] = {x, y, z};
                normalsPtr[writeIndex] = normal;
                writeIndex++;
            }
        }

        faces.resize(RINGS * SIDES * 6);
        int* facesPtr = faces.ptrw();
        writeIndex = 0;

        for (int verticalIt = 0; verticalIt < RINGS; verticalIt++)
        {
            for (int horizontalIt = 0; horizontalIt < SIDES; horizontalIt++)
            {
                int lt = static_cast<int>(horizontalIt + verticalIt * (numVerticesPerRow));
                int rt = static_cast<int>((horizontalIt + 1) + verticalIt * (numVerticesPerRow));

                int lb = static_cast<int>(horizontalIt + (verticalIt + 1) * (numVerticesPerRow));
                int rb = static_cast<int>((horizontalIt + 1) + (verticalIt + 1) * (numVerticesPerRow));

                facesPtr[writeIndex++] = lt;
                facesPtr[writeIndex++] = rt;
                facesPtr[writeIndex++] = lb;

                facesPtr[writeIndex++] = rt;
                facesPtr[writeIndex++] = rb;
                facesPtr[writeIndex++] = lb;
            }
        }
    }

    godot::Array array;
    array.resize(godot::ArrayMesh::ARRAY_MAX);
    array[godot::ArrayMesh::ARRAY_VERTEX] = vertexPositions;
    array[godot::ArrayMesh::ARRAY_NORMAL] = normals;
    array[godot::ArrayMesh::ARRAY_INDEX] = faces;
    godot::ArrayMesh* mesh = memnew(godot::ArrayMesh);
    mesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_TRIANGLES, array);
    torusMeshInstance->set_mesh(mesh);
    float coneHeight = width * 6.0f;
    float coneBottomRadius = width * 2.5f;

    cone->set_scale({coneBottomRadius, coneHeight, coneBottomRadius});
    float angle0 = (RINGS - 1.0f) / RINGS * TWO_PI * completion2;
    float x0 = radius * std::cos(angle0);
    float y0 = radius * std::sin(angle0);
    float angle1 = (RINGS - 0.5f) / RINGS * TWO_PI * completion2;

    cone->set_position({x0, y0, 0});
    godot::Quaternion rotation = godot::Quaternion({0, 0, 1}, angle1);
    cone->set_rotation(rotation.get_euler());
    if (sign < 0)
    {
        cone->rotate({0, 0, 1}, Math_PI);
    }
}

void
avi::scene::ArrowCircleElementNode::update(
    IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto arrowCircle =
        IceInternal::Handle<armarx::viz::data::ElementArrowCircle>::dynamicCast(elementInfo);
    updateAttributes(arrowCircle->interaction.enableFlags,
                     arrowCircle->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(arrowCircle->pose),
                     avi::scene::Converter::getScale(arrowCircle->scale),
                     avi::scene::Converter::getRotation(arrowCircle->pose),
                     avi::scene::Converter::getColor(arrowCircle->color),
                     arrowCircle->completion,
                     arrowCircle->radius,
                     arrowCircle->width);
}

void
avi::scene::ArrowCircleElementNode::setShape()
{
    DETOOLIFY;
    godot::Ref<godot::Shape3D> cylinderShape = cylinderMesh->create_trimesh_shape();
    this->collisionShape->set_shape(cylinderShape);
    collisionShape->set_rotation({godot::Math::deg_to_rad(90.0f), 0, 0});
}

std::vector<godot::Node3D*>
avi::scene::ArrowCircleElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    vec.push_back(cone);
    vec.push_back(torusMeshInstance);
    return vec;
}
