#pragma once

#include "ElementNode.h"

#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/cylinder_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

namespace avi::scene
{
    class ArrowCircleElementNode : public avi::scene::ElementNode
    {
        GDCLASS(ArrowCircleElementNode, ElementNode)

    public:
        static void _bind_methods();

        ArrowCircleElementNode();
        ~ArrowCircleElementNode() override = default;

        static ArrowCircleElementNode*
        newArrowCircleElementNode(avi::connection::SentElementType type,
                                  std::string name,
                                  avi::core::Layer* layer);

        static ArrowCircleElementNode* newArrowCircleElementNode();

        void _ready() override;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        void setColor(godot::Color color);

        void setCompletion(float completion);

        void setRadius(float radius);

        void setWidth(float width);

        void setShape();

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        void setAllAttributes(avi::connection::SentElementType type,
                              std::string name,
                              avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color,
                              float completion,
                              float radius,
                              float width);

    private:
        godot::MeshInstance3D* torusMeshInstance;
        godot::MeshInstance3D* cone;
        godot::StaticBody3D* staticBody;
        godot::CylinderMesh* cylinderMesh;
        godot::CollisionShape3D* collisionShape;
        float completion;
        float radius;
        float width;
        static const int RINGS = 32;
        static const int SIDES = 8;

        bool shapeChecked = false;

        void generateMesh();

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
