#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ArrowElementNode.h"

#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::ArrowElementNode::_bind_methods()
{
    DETOOLIFY;

}

avi::scene::ArrowElementNode::ArrowElementNode() = default;

godot::AABB
avi::scene::ArrowElementNode::getVolume()
{
    godot::AABB correctedAABB = cone->get_aabb();
    correctedAABB.position = staticBody->get_transform().origin;
    correctedAABB.size = correctedAABB.size * 0.5f;


    return correctedAABB;
}

void
avi::scene::ArrowElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::ArrowElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::ArrowElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::ArrowElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::ArrowElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    auto* mat = cast_to<godot::StandardMaterial3D>(cylinder->get_mesh()->surface_get_material(0).ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}

void
avi::scene::ArrowElementNode::setLength(float length)
{
    DETOOLIFY;
    godot::Vector3 scale = cylinder->get_scale();
    scale.y = length;
    cylinder->set_scale(scale);
    cylinder->set_position({0, length / 2, 0});
    cone->set_position({0, length, 0});

    collisionShape->set_position({0, length / 2, 0});
    collisionShape->set_scale(scale);
}

void
avi::scene::ArrowElementNode::setWidth(float width)
{
    DETOOLIFY;
    cone->set_scale({width * 2.5f, width * 6.0f, width * 2.5f});
    cylinder->set_scale({width, std::max(0.0f, cylinder->get_scale().y - width * 6.0f), width});
    cylinder->set_position({0, cylinder->get_scale().y / 2, 0});
    cone->set_position({0, cylinder->get_scale().y, 0});

    collisionShape->set_position({0, cylinder->get_scale().y / 2, 0});
    collisionShape->set_scale({width, std::max(0.0f, cylinder->get_scale().y), width});
}

godot::Vector3
avi::scene::ArrowElementNode::getPosition()
{
    return staticBody->get_position();
}

godot::Vector3
avi::scene::ArrowElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::ArrowElementNode::getScale()
{
    return staticBody->get_scale() * 1000;
}

avi::scene::ArrowElementNode*
avi::scene::ArrowElementNode::newArrowElementNode()
{
    ArrowElementNode* node = memnew(ArrowElementNode);
    node->staticBody = memnew(godot::StaticBody3D);
    node->collisionShape = memnew(godot::CollisionShape3D);
    node->cylinder = memnew(godot::MeshInstance3D);
    node->cone = memnew(godot::MeshInstance3D);
    node->cylinderMesh = memnew(godot::CylinderMesh);

    godot::CylinderMesh* coneMesh = memnew(godot::CylinderMesh);

    node->add_child(node->staticBody);
    node->staticBody->add_child(node->collisionShape);
    node->staticBody->add_child(node->cylinder);
    node->staticBody->add_child(node->cone);

    node->cylinder->set_mesh(node->cylinderMesh);
    node->cone->set_mesh(coneMesh);
    coneMesh->set_top_radius(0);
    coneMesh->set_height(1);

    node->cylinderMesh->set_height(1);
    node->cylinder->set_scale({0.5, 1, 0.5});
    node->cone->set_scale({1, 1, 1});

    godot::StandardMaterial3D* material = memnew(godot::StandardMaterial3D);
    material->set_shading_mode(godot::StandardMaterial3D::SHADING_MODE_PER_PIXEL);

    node->cylinder->get_mesh()->surface_set_material(0, material);
    node->cone->get_mesh()->surface_set_material(0, material);

    return node;
}

void
avi::scene::ArrowElementNode::setAllAttributes(avi::connection::SentElementType type,
                                               std::string name,
                                               avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::ArrowElementNode::updateAttributes(int flag,
                                               std::vector<std::string> interactions,
                                               godot::Vector3 position,
                                               godot::Vector3 scale,
                                               godot::Quaternion rotation,
                                               godot::Color color,
                                               float length,
                                               float width)
{
    setInteractionFlag(flag);
    if (!shapeChecked)
    {
        if (ElementNode::isSelectable())
        {
            setShape();
        }
        shapeChecked = true;
    }
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
    setLength(length);
    setWidth(width);
}

void
avi::scene::ArrowElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto arrow = IceInternal::Handle<armarx::viz::data::ElementArrow>::dynamicCast(elementInfo);
    updateAttributes(arrow->interaction.enableFlags,
                     arrow->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(arrow->pose),
                     avi::scene::Converter::getScale(arrow->scale),
                     avi::scene::Converter::getRotation(arrow->pose),
                     avi::scene::Converter::getColor(arrow->color),
                     arrow->length,
                     arrow->width);
}

godot::Transform3D
avi::scene::ArrowElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

void
avi::scene::ArrowElementNode::setShape()
{
    DETOOLIFY;
    godot::Ref<godot::Shape3D> cylinderShape = cylinderMesh->create_trimesh_shape();
    this->collisionShape->set_shape(cylinderShape);
}

std::vector<godot::Node3D*>
avi::scene::ArrowElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    vec.push_back(cone);
    vec.push_back(cylinder);
    return vec;
}
