#ifndef ARVIZ_INTERACTIVE_PROJECT_SOURCE_ARVIZ_INTERACTIVE_PROJECT_COMPONENTS_ARVIZ_INTERACTIVE_LIB_AVI_SCENE_ARROWELEMENTNODE_H_
#define ARVIZ_INTERACTIVE_PROJECT_SOURCE_ARVIZ_INTERACTIVE_PROJECT_COMPONENTS_ARVIZ_INTERACTIVE_LIB_AVI_SCENE_ARROWELEMENTNODE_H_

#include "ElementNode.h"

#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/cylinder_shape3d.hpp>
#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

namespace avi::scene
{
    class ArrowElementNode : public avi::scene::ElementNode
    {
        GDCLASS(ArrowElementNode, ElementNode)

    public:
        ArrowElementNode();
        ~ArrowElementNode() override = default;

        static ArrowElementNode* newArrowElementNode();

        static void _bind_methods();

        void _ready() override;
        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        void setColor(godot::Color color);

        void setLength(float length);

        void setWidth(float width);

        void setShape();

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        void setAllAttributes(avi::connection::SentElementType type,
                              std::string name,
                              avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color,
                              float length,
                              float width);

    private:
        godot::MeshInstance3D* cylinder;
        godot::MeshInstance3D* cone;
        godot::StaticBody3D* staticBody;
        godot::CylinderMesh* cylinderMesh;
        godot::CollisionShape3D* collisionShape;

        bool shapeChecked = false;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene

#endif //ARVIZ_INTERACTIVE_PROJECT_SOURCE_ARVIZ_INTERACTIVE_PROJECT_COMPONENTS_ARVIZ_INTERACTIVE_LIB_AVI_SCENE_ARROWELEMENTNODE_H_
