#pragma once

#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/node3d.hpp>

#include <arviz_godot/lib/avi/connection/RevisionBuffer.h>
#include <arviz_godot/lib/avi/core/ModelLoader.h>
#include <arviz_godot/lib/avi/scene/CameraSet.h>
#include <arviz_godot/lib/avi/scene/ElementNode.h>
#include <arviz_godot/lib/avi/scene/GraphicsController.h>
#include <arviz_godot/lib/avi/scene/LayerNode.h>

namespace avi::scene
{
    class Converter
    {
    public:
        Converter(godot::Node3D* sceneRoot,
                  avi::connection::RevisionBuffer& buffer,
                  avi::scene::GraphicsController* graphicsController);

        struct ChangedLayers {
            std::map<LayerName, avi::scene::LayerNode*> layers;
            bool clearScene;
        };

        ChangedLayers apply();

        void clear();
        void clearCaches();

        static godot::Vector3 getPosition(armarx::viz::data::GlobalPose pose);

        static godot::Vector3 getScale(armarx::Vector3f vec);

        static godot::Quaternion getRotation(armarx::viz::data::GlobalPose pose);

        static godot::Color getColor(armarx::viz::data::Color);

        static godot::Vector3 getNormal(armarx::Vector3f A, armarx::Vector3f B, armarx::Vector3f C);



    private:
        void updateLayer(std::optional<avi::scene::LayerNode*> optLayer,
                         armarx::viz::data::LayerUpdate update);

        avi::scene::ElementNode*
        createElement(avi::scene::LayerNode* layer,
                      IceInternal::Handle<armarx::viz::data::Element> elementInfo);

        avi::scene::ElementNode*
        updateElement(avi::scene::LayerNode* layer,
                      avi::scene::ElementNode* element,
                      IceInternal::Handle<armarx::viz::data::Element> elementInfo);

        godot::Node3D* sceneRoot;
        avi::connection::RevisionBuffer& buffer;
        avi::core::ModelLoader modelLoader;
        std::map<LayerName, avi::scene::LayerNode*> layers{};
        avi::scene::GraphicsController* graphicsController;

        static avi::connection::SentElementType
        getType(IceInternal::Handle<armarx::viz::data::Element> element);

        static float adjustSize(float size);

        static armarx::Vector3f adjustSizeVector(armarx::Vector3f sizevec);

        static godot::ArrayMesh* getNonEmptyArrayMesh();
    };
} // namespace avi::scene
