#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "CustomMeshElementNode.h"

#include <godot_cpp/classes/mesh_data_tool.hpp>

#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::CustomMeshElementNode::_bind_methods()
{
    DETOOLIFY;

}

void
avi::scene::CustomMeshElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::CustomMeshElementNode::update(
    IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto meshEl = IceInternal::Handle<armarx::viz::data::ElementMesh>::dynamicCast(elementInfo);
    auto* mesh = godot::Object::cast_to<godot::ArrayMesh>(getMesh());

    int storageSize = meshEl->faces.size() * 6;

    godot::PackedVector3Array coords;
    coords.resize(storageSize);
    godot::Vector3* coordsPtr = coords.ptrw();

    godot::PackedColorArray colors;
    colors.resize(storageSize);
    godot::Color* colorsPtr = colors.ptrw();

    godot::PackedVector3Array normals;
    normals.resize(storageSize);
    godot::Vector3* normalsPtr = normals.ptrw();

    size_t writeIndex = 0;

    for (auto face : meshEl->faces)
    {
        auto putVertex = [&](int vertexIdx, int colorIdx, godot::Vector3 normal)
        {
            armarx::Vector3f vertex = meshEl->vertices.at(vertexIdx);
            coordsPtr[writeIndex] = godot::Vector3(vertex.e0, vertex.e1, vertex.e2);
            colorsPtr[writeIndex] = avi::scene::Converter::getColor(meshEl->colors.at(colorIdx));
            normalsPtr[writeIndex] = normal;

            writeIndex++;
        };

        godot::Vector3 normal1 = avi::scene::Converter::getNormal(meshEl->vertices.at(face.v0),
                                                                  meshEl->vertices.at(face.v2),
                                                                  meshEl->vertices.at(face.v1));

        putVertex(face.v0, face.c0, normal1);
        putVertex(face.v1, face.c1, normal1);
        putVertex(face.v2, face.c2, normal1);

        godot::Vector3 normal2 = avi::scene::Converter::getNormal(meshEl->vertices.at(face.v0),
                                                                  meshEl->vertices.at(face.v1),
                                                                  meshEl->vertices.at(face.v2));

        putVertex(face.v0, face.c0, normal2);
        putVertex(face.v2, face.c2, normal2);
        putVertex(face.v1, face.c1, normal2);
    }

    godot::Array arr{};
    arr.resize(godot::ArrayMesh::ARRAY_MAX);
    arr[godot::ArrayMesh::ARRAY_VERTEX] = coords;
    arr[godot::ArrayMesh::ARRAY_COLOR] = colors;
    arr[godot::ArrayMesh::ARRAY_NORMAL] = normals;
    mesh->clear_surfaces();
    mesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_TRIANGLES, arr);
    mesh->surface_set_material(0, memnew(godot::StandardMaterial3D));
    updateAttributes(meshEl->interaction.enableFlags,
                     meshEl->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(meshEl->pose),
                     avi::scene::Converter::getScale(meshEl->scale),
                     avi::scene::Converter::getRotation(meshEl->pose),
                     avi::scene::Converter::getColor(meshEl->color));
    useVertexColor(true);

    if (first)
    {
        setShape(mesh);
        first = false;
    }
}

avi::scene::CustomMeshElementNode*
avi::scene::CustomMeshElementNode::newCustomMeshElementNode(godot::Mesh* mesh,
                                                            avi::connection::SentElementType type,
                                                            std::string name,
                                                            avi::core::Layer* layer)
{
    CustomMeshElementNode* node = memnew(CustomMeshElementNode);
    node->setStartAttributes(mesh, type, name, layer);

    return node;
}
