#pragma once

#include <memory>
#include <optional>

#include <Ice/Handle.h>

#include <RobotAPI/interface/ArViz/Elements.h>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/area3d.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <arviz_godot/lib/avi/core/Element.h>
#include <arviz_godot/lib/avi/scene/DrawStyle.h>

namespace avi::scene
{
    class ElementNode : public godot::Node3D, public avi::core::Element
    {
        GDCLASS(ElementNode, Node3D)

    public:
        static void _bind_methods();

        ElementNode();
        ~ElementNode() override = default;

        void _notification(int what);

        avi::connection::SentElementType getType() override;

        std::string getName() override;

        avi::core::Layer* getLayer() override;

        int getFlag();

        bool isSelectable() override;

        bool isManipulable() override;

        bool hasContextMenu() override;

        bool translationX();

        bool translationY();

        bool translationZ();

        bool translationLocal();

        bool rotationX();

        bool rotationY();

        bool rotationZ();

        bool rotationLocal();

        bool scalingX();

        bool scalingY();

        bool scalingZ();

        bool scalingLocal();

        bool transformHide();

        std::vector<std::string> getInteractions() override;

        void setType(avi::connection::SentElementType type);

        void setDeselectCallback(std::function<void()> callback) override;

        void setName(const std::string& name);

        void setLayer(avi::core::Layer* layer);

        void setInteractions(std::vector<std::string> interactions);

        void setInteractionFlag(int flag);

        void setMeshVisible(bool visible);

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void createGhostElement(godot::Transform3D transform);

        void deselect();

        virtual void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo);

    private:
        avi::connection::SentElementType type;
        std::string name;
        avi::core::Layer* layer;
        int interactionFlags;
        std::vector<std::string> interactions;
        std::optional<std::function<void()>> deselectCallback;

        godot::Node3D* ghostElement{nullptr};

    protected:
        virtual std::vector<godot::Node3D*> getSpatial();
    };

    class ModelDrawStyle
    {
    public:
        ModelDrawStyle();
        explicit ModelDrawStyle(int drawStyle);

        enum Model
        {
            VISUALIZATION,
            COLLISION
        };

        Model getModel() const;
        bool isVisualization() const;
        bool isCollision() const;
        bool isColorOverriden() const;

        avi::scene::DrawStyle getSceneDrawStyle();

        bool operator==(ModelDrawStyle const& other) const;
        bool operator!=(ModelDrawStyle const& other) const;

    private:
        Model model{};
        bool overrideColor{};
        int armarXCodedValue;
    };
} // namespace avi::scene
