#pragma once

#include <map>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <arviz_godot/lib/avi/core/Layer.h>
#include <arviz_godot/lib/avi/scene/ElementNode.h>

namespace avi::scene
{
    struct LayerName
    {
        std::string component;
        std::string name;

        bool operator==(const LayerName& other) const;

        bool operator<(const LayerName& other) const;
    };

    class LayerNode : public godot::Node3D, public avi::core::Layer
    {
        GDCLASS(LayerNode, Node3D)

    public:
        static void _bind_methods();

        LayerNode();
        ~LayerNode() override = default;

        bool isVisible() const override;

        void setVisibility(bool isVisible);

        void setName(const std::string& component, const std::string& name);

        const std::string& getName() override;

        const std::string& getComponent() override;

        void addElement(avi::scene::ElementNode* node);

        avi::scene::ElementNode* getElement(std::string name);

        void removeElement(std::string name);

        void saveRemove();

        godot::AABB getAABB();

        void _notification(int what);

    private:
        std::string component;
        std::string name;
        bool visible;
        std::map<std::string, avi::scene::ElementNode*> elements;
    };
} // namespace avi::scene
