#pragma once

#include "MeshElementNode.h"

namespace godot
{
    class CylinderMesh;
    class StandardMaterial3D;
} // namespace godot

namespace avi::scene
{
    class LineElementNode : public avi::scene::ElementNode
    {
        GDCLASS(LineElementNode, ElementNode)

    public:
        static void _bind_methods();

        LineElementNode() = default;
        ~LineElementNode() override = default;

        void _ready() override;

        static LineElementNode* newLineElementNode(std::string name, avi::core::Layer* layer);

        void setStartAttributes(std::string name, avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;
        std::vector<godot::Node3D*> getSpatial() override;

    private:
        void updateRadius(float newRadius);
        void updateLine(godot::Vector3 from, godot::Vector3 to);

        void setPosition(godot::Vector3 position);
        void setRotation(godot::Quaternion rotation);
        void setScale(godot::Vector3 scale);
        void setColor(godot::Color color);

    private:
        float radius = 0.0;
        godot::Vector3 a = godot::Vector3(0, 0, 0);
        godot::Vector3 b = godot::Vector3(0, 0, 0);

        godot::StaticBody3D* staticBody;
        godot::CollisionShape3D* collisionLine;
        godot::StandardMaterial3D* material;

        godot::MeshInstance3D* start;
        godot::MeshInstance3D* end;
        godot::MeshInstance3D* line;
    };
} // namespace avi::scene
