#pragma once

#include "ElementNode.h"

#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

#include <arviz_godot/lib/avi/core/Element.h>

namespace avi::scene
{
    class MeshElementNode : public avi::scene::ElementNode
    {
        GDCLASS(MeshElementNode, ElementNode)

    public:
        MeshElementNode();
        ~MeshElementNode() override;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        void setColor(godot::Color color);

        void setMesh(godot::Mesh* mesh);

        void setShape(godot::Mesh* mesh);
        void setShape(godot::Shape3D* shape);

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        godot::MeshInstance3D* getMeshInstance();
        godot::CollisionShape3D* getCollisionShape();

        void setStartAttributes(godot::Mesh* mesh,
                                avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

        godot::Mesh* getMesh();

        void useVertexColor(bool isUsed);

        void setTwoSided();

    private:
        godot::StaticBody3D* staticBody;
        godot::CollisionShape3D* collisionShape;
        godot::MeshInstance3D* meshInstance;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
