#pragma once

#include "ElementNode.h"

#include <godot_cpp/classes/shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

namespace avi::scene
{
    class MultiMeshElementNode : public avi::scene::ElementNode
    {
        GDCLASS(MultiMeshElementNode, avi::scene::ElementNode)

    public:
        MultiMeshElementNode();
        ~MultiMeshElementNode() override;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void setPosition(godot::Vector3 position, size_t index);

        void setScale(godot::Vector3 scale, size_t index);

        void setRotation(godot::Quaternion rotation, size_t index);

        void setColor(godot::Color color, size_t index);

        void setMesh(std::vector<godot::Mesh*> meshes);

        void setShape();

        std::vector<godot::Vector3> getPosition();

        std::vector<godot::Vector3> getRotation();

        std::vector<godot::Vector3> getScale();

        void setStartAttributes(std::vector<godot::Mesh*> meshes,
                                avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

        godot::Mesh* getMesh(int index);

        void setTwoSided(int index);

        void setVisibility(int index, bool visible);

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

    private:
        std::vector<godot::MeshInstance3D*> meshInstances;
        godot::StaticBody3D* staticBody;
        std::vector<godot::CollisionShape3D*> collisionShapes;

        bool shapeChecked = false;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
