#pragma once

#include <vector>

#include "ElementNode.h"

#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>
#include <godot_cpp/classes/shape3d.hpp>
#include <godot_cpp/classes/texture2d.hpp>

#include <arviz_godot/lib/avi/core/ModelLoader.h>

namespace avi::scene
{
    class ObjectElementNode : public avi::scene::ElementNode
    {
        GDCLASS(ObjectElementNode, ElementNode)

    public:
        static void _bind_methods();

        ObjectElementNode();
        ~ObjectElementNode() override = default;

        static ObjectElementNode* newObjectElementNode(avi::core::ModelLoader* loader,
                                                       avi::connection::SentElementType type,
                                                       std::string name,
                                                       avi::core::Layer* layer);

        void _ready() override;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        void setShape(VirtualRobot::ManipulationObjectPtr visualization);

        void setStartAttributes(avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

        void updateAttributes(std::string fileName,
                              int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color,
                              ModelDrawStyle drawStyle);

        void setOverrideColor(godot::Color color, ModelDrawStyle drawStyle);

    private:
        avi::core::ModelLoader* loader{nullptr};

        std::string fileName;

        godot::StaticBody3D* staticBody{nullptr};

        godot::MeshInstance3D* meshInstance{nullptr};
        godot::CollisionShape3D* collisionShape{nullptr};

        ModelDrawStyle currentDrawStyle{};
        godot::Color currentColor{};

        bool shapeChecked = false;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
