#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "OrientationIndicator.h"

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/input_event.hpp>
#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/world3d.hpp>

#include <arviz_godot/lib/avi/scene/CameraControlScheme.h>
#include <arviz_godot/lib/avi/scene/PlayfulControlScheme.h>
#include <arviz_godot/lib/avi/scene/TechnicalControlScheme.h>

void
avi::scene::OrientationIndicator::_bind_methods()
{
    DETOOLIFY;
}

avi::scene::OrientationIndicator::OrientationIndicator() = default;
void
avi::scene::OrientationIndicator::_ready()
{
    DETOOLIFY;
    pose = get_node<godot::StaticBody3D>("Pose");
    godot::Vector2 viewportSize = godot::Vector2(200, 200);
    this->set_size(viewportSize);
}

void
avi::scene::OrientationIndicator::_process(double)
{
    DETOOLIFY;
    CameraControlScheme* scheme = viewCamera->getCurrentScheme();
    godot::Transform3D quaternionTransform;

    pose->set_rotation_degrees(scene->getOrientation());
    if (dynamic_cast<TechnicalControlScheme*>(scheme))
    {
        quaternionTransform =
            scheme->getRotationBasis() * pose->get_global_transform().basis.orthonormalized();
    }
    else
    {
        quaternionTransform = scheme->getRotationBasis().inverse() *
                              pose->get_global_transform().basis.orthonormalized();
    }

    quaternionTransform.origin = pose->get_global_transform().origin;
    quaternionTransform.scale(godot::Vector3(0.09, 0.09, 0.09));
    pose->set_global_transform(quaternionTransform);
}

void
avi::scene::OrientationIndicator::setup(avi::scene::VisualizedScene* scene,
                                        avi::scene::ViewCamera* viewCamera,
                                        avi::core::EventInvokingSystem& events)
{
    DETOOLIFY;
    this->viewCamera = viewCamera;
    this->camera = get_node<godot::Camera3D>("Camera");

    points.push_back({get_node<godot::Node3D>("Pose/X-Point"), CameraControlScheme::Axis::xAxis});
    points.push_back(
        {get_node<godot::Node3D>("Pose/X-Point-Negative"), CameraControlScheme::Axis::xAxis});
    points.push_back({get_node<godot::Node3D>("Pose/Y-Point"), CameraControlScheme::Axis::yAxis});
    points.push_back(
        {get_node<godot::Node3D>("Pose/Y-Point-Negative"), CameraControlScheme::Axis::yAxis});
    points.push_back({get_node<godot::Node3D>("Pose/Z-Point"), CameraControlScheme::Axis::zAxis});
    points.push_back(
        {get_node<godot::Node3D>("Pose/Z-Point-Negative"), CameraControlScheme::Axis::zAxis});

    this->scene = scene;
    this->events = &events;
}

void
avi::scene::OrientationIndicator::_input(godot::Ref<godot::InputEvent> const& event)
{
    DETOOLIFY;
    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent)
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == godot::MOUSE_BUTTON_LEFT && event->is_pressed())
        {
            selectElement();
        }
    }
}

void
avi::scene::OrientationIndicator::selectElement()
{
    DETOOLIFY;
    godot::Vector2 mouse_position = get_mouse_position();
    std::vector<OrientationIndicatorPoint> pointsInDistance{};

    for (OrientationIndicatorPoint point : points)
    {
        float distance = (mouse_position -
                          camera->unproject_position(point.spatial->get_global_transform().origin))
                             .length();
        if (distance < 15)
        {
            pointsInDistance.push_back(point);
        }
    }

    if (pointsInDistance.size() == 1)
    {
        viewCamera->placeOnAxis(pointsInDistance[0].spatial->get_global_transform().origin);
    }

    if (pointsInDistance.size() == 2)
    {
        if (pointsInDistance[0].axis == pointsInDistance[1].axis)
        {
            if (pointsInDistance[0].spatial->get_global_transform().origin.z <
                pointsInDistance[1].spatial->get_global_transform().origin.z)
            {
                viewCamera->invertAxis(
                    pointsInDistance[0].spatial->get_global_transform().origin.normalized());
            }
            else
            {
                viewCamera->invertAxis(
                    pointsInDistance[1].spatial->get_global_transform().origin.normalized());
            }
        }
        else
        {
            if (pointsInDistance[0].spatial->get_global_transform().origin.z >
                pointsInDistance[1].spatial->get_global_transform().origin.z)
            {
                viewCamera->placeOnAxis(pointsInDistance[0].spatial->get_global_transform().origin);
            }
            else
            {
                viewCamera->placeOnAxis(pointsInDistance[1].spatial->get_global_transform().origin);
            }
        }
    }
}
