#pragma once

#include "ElementNode.h"

#include <godot_cpp/classes/standard_material3d.hpp>

namespace godot
{
    class StaticBody3D;
    class CollisionShape3D;
    class SpatialMaterial;
    class MultiMesh;
} // namespace godot

namespace avi::scene
{
    class PathElementNode : public avi::scene::ElementNode
    {
        GDCLASS(PathElementNode, ElementNode)

    public:
        static void _bind_methods();

        PathElementNode() = default;
        ~PathElementNode() override = default;

        void _ready() override;

        static avi::scene::PathElementNode* newPathElementNode(std::string name,
                                                               avi::core::Layer* layer);

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setStartAttributes(std::string name, avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;
        std::vector<godot::Node3D*> getSpatial() override;

    private:
        void buildPath();

        void buildLines();
        void buildSpheres();

        void setPosition(godot::Vector3 position);
        void setRotation(godot::Quaternion rotation);
        void setScale(godot::Vector3 scale);
        void setColor(godot::Color color);

    private:
        float radius = 0.0;
        std::vector<godot::Vector3> points = {};

        std::vector<godot::Node3D*> spatials = {};

        godot::StaticBody3D* staticBody;
        godot::StandardMaterial3D* material;

        godot::MultiMesh* meshLine;
        godot::MultiMesh* meshSphere;

        godot::PackedFloat32Array lineData;
        godot::PackedFloat32Array sphereData;

        godot::Vector3 maxPoint;
        godot::Vector3 minPoint;

        std::vector<godot::CollisionShape3D*> shapes = {};
    };
} // namespace avi::scene
