#pragma once

#include "CameraControlScheme.h"
#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <arviz_godot/lib/avi/scene/ViewCamera.h>
#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

namespace avi::scene
{
    class PlayfulControlScheme : public godot::Node3D, public avi::scene::CameraControlScheme
    {
        GDCLASS(PlayfulControlScheme, Node3D)

    public:
        PlayfulControlScheme();
        ~PlayfulControlScheme() override = default;

        static void _bind_methods();

        void _ready() override;

        void _process(double delta) override;

        void _unhandled_input(godot::Ref<godot::InputEvent> const& event) override;

    public:
        void setup(avi::core::SettingsProvider& settings,
                   avi::scene::VisualizedScene& scene,
                   bool gravity);

        void takeOver(avi::scene::ViewCamera* viewCamera, godot::Transform3D head) override;

        godot::Transform3D release() override;

        void reset() override;

        void focus(godot::AABB box) override;

        void focus(godot::Vector3 point, avi::core::Element* target) override;

        void placeOnAxis(godot::Vector3 fromVector) override;

        void invertAxis(godot::Vector3 fromVector) override;

        void placeOnAxisNumpad(Axis axis, AxisSign sign) override;

        void rotateFixed(RotationDirection direction, HotkeyRotation hotkey) override;

        godot::Basis getRotationBasis() override;

        void setupControlAdvertisements(avi::core::ControlAdvertiser& advertiser) override;

        void updateControlAdvertisements();

    private:
        bool isActive();

        void startTravel(godot::Transform3D target, float duration);

        void doTravel(float delta);

    private:
        struct Travel
        {
            godot::Transform3D startTransform{};
            godot::Transform3D endTransform{};
            godot::Transform3D endTransformHead{};

            float duration{};
            float remainingTime{};
        };

        Travel travel{};
        bool isTraveling{false};

        bool gravity = false;

        godot::Node3D* attachment;
        godot::Node3D* head;
        avi::scene::ViewCamera* viewCamera;

        float cameraStartSize = 1.0f;
        float cameraEndSize = 1.0f;
        bool updateSize = false;

        std::vector<bool> inputState;
        std::vector<const char*> inputKeys;

        enum InputStateIndex
        {
            INPUT_FORWARD,
            INPUT_BACKWARD,
            INPUT_LEFT,
            INPUT_RIGHT,
            INPUT_UP,
            INPUT_DOWN,
            INPUT_ROLL_LEFT,
            INPUT_ROLL_RIGHT,
            INPUT_SPEED_DOWN,
            INPUT_SPEED_UP,
            MAX_INPUT_INDEX
        };

        bool switchX = false;
        bool switchY = false;
        bool switchZ = false;

        bool rotatePressed = false;
        bool continuousEnabled = true;
        godot::Vector2 pressedRotation;

        avi::core::SettingsProvider* settings;
        avi::scene::VisualizedScene* scene;

        godot::Vector2 mouseDeltaAccumulator {};
        godot::Vector2 smoothMouseDelta {};
        float forwardMovement = 0.0f;
        float speedModifier;

        godot::Vector3 smoothMovement = {};
        float smoothRoll = 0.0f;

        core::ControlAdvertiser* advertiser = nullptr;
        size_t adID;
        size_t adIDOrthographic;
    };
} // namespace avi::scene
