#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "PointCloudElementNode.h"

#include <godot_cpp/classes/box_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/point_mesh.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::PointCloudElementNode::_bind_methods()
{
    DETOOLIFY;

}

avi::scene::PointCloudElementNode::PointCloudElementNode()
{
    DETOOLIFY;
}

void
avi::scene::PointCloudElementNode::_ready()
{
    DETOOLIFY;
}

avi::scene::PointCloudElementNode*
avi::scene::PointCloudElementNode::newPointCloudElementNode(avi::connection::SentElementType type,
                                                            std::string name,
                                                            avi::core::Layer* layer)
{
    PointCloudElementNode* node = memnew(PointCloudElementNode);
    node->staticBody = memnew(godot::StaticBody3D);
    node->collisionShape = memnew(godot::CollisionShape3D);
    node->meshInstance = memnew(godot::MeshInstance3D);

    node->mesh = memnew(godot::ArrayMesh);

    node->staticBody->add_child(node->collisionShape);
    node->staticBody->add_child(node->meshInstance);
    node->add_child(node->staticBody);

    node->meshInstance->set_mesh(node->mesh);

    node->material = godot::Ref(memnew(godot::StandardMaterial3D));
    node->material->set_flag(godot::StandardMaterial3D::FLAG_USE_POINT_SIZE, true);
    node->material->set_flag(godot::StandardMaterial3D::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);
    node->material->set_shading_mode(godot::StandardMaterial3D::SHADING_MODE_UNSHADED);

    node->setStartAttributes(type, name, layer);
    return node;
}

void
avi::scene::PointCloudElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    this->position = position;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::PointCloudElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::PointCloudElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    this->color = color;
}

void
avi::scene::PointCloudElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

godot::Vector3
avi::scene::PointCloudElementNode::getPosition()
{
    return position;
}

godot::Vector3
avi::scene::PointCloudElementNode::getScale()
{
    return staticBody->get_scale();
}

void
avi::scene::PointCloudElementNode::setStartAttributes(avi::connection::SentElementType type,
                                                      std::string name,
                                                      avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::PointCloudElementNode::updateAttributes(int flag,
                                                    std::vector<std::string> interactions,
                                                    godot::Vector3 position,
                                                    godot::Vector3 scale,
                                                    godot::Quaternion rotation,
                                                    godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    if (ElementNode::isSelectable())
    {
        setShape();
    }

    setInteractions(interactions);
    setPosition(position);
    setColor(color);
    setRotation(rotation);
    setScale(scale);
}

void
avi::scene::PointCloudElementNode::drawPointCloud(ulong pointCloudSize,
                                                  const armarx::viz::data::ColoredPoint* pointPtr)
{
    DETOOLIFY;
    int requiredSize = static_cast<int>(pointCloudSize);

    if (pointData.size() != requiredSize)
    {
        pointData.resize(requiredSize);
    }
    if (colorData.size() != requiredSize)
    {
        colorData.resize(requiredSize);
    }

    godot::Vector3* positionDataBase = pointData.ptrw();
    godot::Color* colorDataBase = colorData.ptrw();

    for (ulong pointIndex = 0; pointIndex < pointCloudSize; pointIndex++)
    {
        const armarx::viz::data::ColoredPoint& point = pointPtr[pointIndex];

        auto getColor = [](const armarx::viz::data::ColoredPoint& point) -> godot::Color
        {
            godot::Color pointColor;

            pointColor.r =
                static_cast<float>(point.color.r) / std::numeric_limits<Ice::Byte>::max();
            pointColor.g =
                static_cast<float>(point.color.g) / std::numeric_limits<Ice::Byte>::max();
            pointColor.b =
                static_cast<float>(point.color.b) / std::numeric_limits<Ice::Byte>::max();
            pointColor.a =
                static_cast<float>(point.color.a) / std::numeric_limits<Ice::Byte>::max();

            return pointColor;
        };

        positionDataBase[pointIndex] = godot::Vector3(point.x, point.y, point.z);
        colorDataBase[pointIndex] = getColor(point);

        minPoint = {std::min(minPoint.x, point.x),
                    std::min(minPoint.y, point.y),
                    std::min(minPoint.z, point.z)};
        maxPoint = {std::max(maxPoint.x, point.x),
                    std::max(maxPoint.y, point.y),
                    std::max(maxPoint.z, point.z)};
    }

    godot::Array array;
    array.resize(godot::ArrayMesh::ARRAY_MAX);

    array[godot::ArrayMesh::ARRAY_VERTEX] = pointData;
    array[godot::ArrayMesh::ARRAY_COLOR] = colorData;

    if (mesh->get_surface_count() > 0)
    {
        mesh->clear_surfaces();
    }

    mesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_POINTS, array);
    mesh->surface_set_material(0, material);
}

godot::AABB
avi::scene::PointCloudElementNode::getVolume()
{
    godot::Vector3 position = minPoint;
    godot::Vector3 size = maxPoint - minPoint;

    return {position / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR,
            size / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR};
}

void
avi::scene::PointCloudElementNode::update(
    IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto pointCloud =
        IceInternal::Handle<armarx::viz::data::ElementPointCloud>::dynamicCast(elementInfo);

    material->set_point_size(pointCloud->pointSizeInPixels);

    ulong pointCloudSize = pointCloud->points.size() / sizeof(armarx::viz::data::ColoredPoint);
    const auto* pointPtr =
        reinterpret_cast<armarx::viz::data::ColoredPoint const*>(pointCloud->points.data());

    auto* pointCloudNode = godot::Object::cast_to<avi::scene::PointCloudElementNode>(this);

    pointCloudNode->updateAttributes(pointCloud->interaction.enableFlags,
                                     pointCloud->interaction.contextMenuOptions,
                                     avi::scene::Converter::getPosition(pointCloud->pose),
                                     avi::scene::Converter::getScale(pointCloud->scale),
                                     avi::scene::Converter::getRotation(pointCloud->pose),
                                     avi::scene::Converter::getColor(pointCloud->color));

    pointCloudNode->drawPointCloud(pointCloudSize, pointPtr);
}

godot::Transform3D
avi::scene::PointCloudElementNode::getTransform()
{
    return staticBody->get_transform();
}

void
avi::scene::PointCloudElementNode::setShape()
{
    DETOOLIFY;
    if (!shapeExists)
    {
        godot::Ref<godot::BoxShape3D> shape = memnew(godot::BoxShape3D);
        collisionShape->set_shape(shape);

        shapeExists = true;
    }

    godot::AABB volume = getVolume();

    if (volume != shapeVolume)
    {
        shapeVolume = volume;

        volume.position *= avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;
        volume.size *= avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;

        collisionShape->set_scale(volume.size * 0.5f);
        collisionShape->set_position(volume.position + volume.size * 0.5f);
    }
}

std::vector<godot::Node3D*>
avi::scene::PointCloudElementNode::getSpatial()
{
    return {};
}
