#pragma once

#include <RobotAPI/interface/ArViz/Elements.h>

#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>
#include <godot_cpp/classes/texture.hpp>

#include <arviz_godot/lib/avi/scene/ElementNode.h>
#include <arviz_godot/lib/avi/scene/GraphicsController.h>

namespace avi::scene
{

    class PointCloudElementNode : public avi::scene::ElementNode
    {
        GDCLASS(PointCloudElementNode, ElementNode)

    public:
        static void _bind_methods();

        PointCloudElementNode();
        ~PointCloudElementNode() override = default;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void _ready() override;

        void drawPointCloud(ulong pointCloudSize, armarx::viz::data::ColoredPoint const* pointPtr);

        static PointCloudElementNode*
        newPointCloudElementNode(avi::connection::SentElementType type,
                                 std::string name,
                                 avi::core::Layer* layer);

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setColor(godot::Color color);

        void setRotation(godot::Quaternion rotation);

        void setShape();

        godot::Vector3 getPosition();

        godot::Vector3 getScale();

        void setStartAttributes(avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

    private:
        godot::MeshInstance3D* meshInstance;
        godot::ArrayMesh* mesh;

        godot::PackedVector3Array pointData{};
        godot::PackedColorArray colorData{};

        godot::Ref<godot::StandardMaterial3D> material;
        godot::Vector3 position;
        godot::Vector3 scale;
        godot::Color color;

        godot::StaticBody3D* staticBody;
        godot::CollisionShape3D* collisionShape;

        godot::Vector3 minPoint{};
        godot::Vector3 maxPoint{};

        bool shapeExists = false;
        godot::AABB shapeVolume;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
