#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "PolygonElementNode.h"

#include "Converter.h"

#include <godot_cpp/classes/surface_tool.hpp>

void
avi::scene::PolygonElementNode::_bind_methods()
{
    DETOOLIFY;

}

void
avi::scene::PolygonElementNode::_ready()
{
    DETOOLIFY;

}

void
avi::scene::PolygonElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;

    auto polygon = IceInternal::Handle<armarx::viz::data::ElementPolygon>::dynamicCast(elementInfo);

    auto* mesh1 = godot::Object::cast_to<godot::ArrayMesh>(getMesh(0));
    mesh1->clear_surfaces();

    auto* mesh2 = godot::Object::cast_to<godot::ArrayMesh>(getMesh(1));
    mesh2->clear_surfaces();

    godot::SurfaceTool lines;
    lines.begin(godot::Mesh::PRIMITIVE_LINE_STRIP);

    godot::PackedVector3Array coords;
    godot::PackedVector3Array normals;

    godot::Vector3 normal = avi::scene::Converter::getNormal(
        polygon->points.at(0), polygon->points.at(1), polygon->points.at(2));
    std::optional<godot::Vector3> first;

    for (auto& vertex : polygon->points)
    {
        godot::Vector3 position = godot::Vector3(vertex.e0, vertex.e1, vertex.e2);

        coords.push_back(position);
        normals.push_back(normal);

        first = first.value_or(position);
        lines.add_vertex(position);
    }

    lines.add_vertex(first.value_or(godot::Vector3(1, 1, 1)));

    godot::SurfaceTool polygonSurface;
    polygonSurface.begin(godot::Mesh::PRIMITIVE_TRIANGLES);
    polygonSurface.add_triangle_fan(coords, godot::PackedVector2Array(), godot::PackedColorArray(), godot::PackedVector2Array(), normals);

    mesh1->add_surface_from_arrays(godot::Mesh::PRIMITIVE_TRIANGLES, polygonSurface.commit_to_arrays());
    mesh1->surface_set_material(0, memnew(godot::StandardMaterial3D));

    mesh2->add_surface_from_arrays(godot::Mesh::PRIMITIVE_LINE_STRIP, lines.commit_to_arrays());
    mesh2->surface_set_material(0, memnew(godot::StandardMaterial3D));

    updateAttributes(polygon->interaction.enableFlags,
                     polygon->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(polygon->pose),
                     avi::scene::Converter::getScale(polygon->scale),
                     avi::scene::Converter::getRotation(polygon->pose),
                     avi::scene::Converter::getColor(polygon->color));

    setTwoSided(0);
    setColor(avi::scene::Converter::getColor(polygon->lineColor), 1);
    if (polygon->lineWidth <= 0)
    {
        setVisibility(1, false);
    }
    else
    {
        setVisibility(1, true);
    }
}

avi::scene::PolygonElementNode*
avi::scene::PolygonElementNode::newPolygonElementNode(std::vector<godot::Mesh*> mesh,
                                                      avi::connection::SentElementType type,
                                                      std::string name,
                                                      avi::core::Layer* layer)
{
    PolygonElementNode* node = memnew(PolygonElementNode);
    node->setStartAttributes(mesh, type, name, layer);
    return node;
}
