#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "PoseElementNode.h"

#include <iostream>

#include <godot_cpp/classes/box_mesh.hpp>
#include <godot_cpp/classes/font.hpp>
#include <godot_cpp/classes/resource_loader.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

godot::Mesh* avi::scene::PoseElementNode::mesh = {};

void
avi::scene::PoseElementNode::_bind_methods()
{
    DETOOLIFY;
}


avi::scene::PoseElementNode::PoseElementNode() = default;

godot::AABB
avi::scene::PoseElementNode::getVolume()
{
    godot::AABB AABB = poseMeshInstance->get_aabb();
    AABB.position = staticBody->get_transform().origin;
    AABB.size = AABB.size * staticBody->get_scale();

    return AABB;
}

void
avi::scene::PoseElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);

    label->set_position({5, 5, 2.5});
}

void
avi::scene::PoseElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::PoseElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::PoseElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
}

godot::Vector3
avi::scene::PoseElementNode::getPosition()
{
    return staticBody->get_position();
}

godot::Vector3
avi::scene::PoseElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::PoseElementNode::getScale()
{
    return staticBody->get_scale();
}

avi::scene::PoseElementNode*
avi::scene::PoseElementNode::newPoseElementNode(avi::connection::SentElementType type,
                                                std::string name,
                                                avi::core::Layer* layer)
{
    PoseElementNode* node = newPoseElementNode();
    node->setStartAttributes(type, name, layer);

    return node;
}

avi::scene::PoseElementNode*
avi::scene::PoseElementNode::newPoseElementNode()
{
    godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
    godot::Ref<godot::Font> font = loader->load("res://fonts/font.tres");

    PoseElementNode* node = memnew(PoseElementNode);

    node->staticBody = memnew(godot::StaticBody3D);
    node->collisionShape = memnew(godot::CollisionShape3D);
    node->poseMeshInstance = memnew(godot::MeshInstance3D);

    node->label = memnew(godot::Label3D);
    node->label->set_font(font);
    node->label->set_font_size(128);
    node->label->set_pixel_size(node->textDownscaling);
    node->label->set_modulate(godot::Color());
    node->label->set_horizontal_alignment(godot::HORIZONTAL_ALIGNMENT_LEFT);
    node->label->set_vertical_alignment(godot::VERTICAL_ALIGNMENT_BOTTOM);

    node->staticBody->add_child(node->collisionShape);
    node->staticBody->add_child(node->poseMeshInstance);
    node->staticBody->add_child(node->label);

    node->add_child(node->staticBody);

    return node;
}

void
avi::scene::PoseElementNode::setStartAttributes(avi::connection::SentElementType type,
                                                std::string name,
                                                avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);

    if (mesh == nullptr)
    {
        godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
        mesh = godot::Ref<godot::Mesh>(loader->load("res://models/pose.obj")).ptr();
    }

    poseMeshInstance->set_mesh(mesh);

    label->set_text(name.c_str());
}

void
avi::scene::PoseElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quaternion rotation,
                                              godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);

    if (!shapeChecked)
    {
        setShape();
        shapeChecked = true;
    }

    setInteractions(interactions);
    setPosition(position);
    setColor(color);
    setRotation(rotation);
    setScale(scale);
}

void
avi::scene::PoseElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto pose = IceInternal::Handle<armarx::viz::data::ElementPose>::dynamicCast(elementInfo);
    updateAttributes(pose->interaction.enableFlags,
                     pose->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(pose->pose),
                     avi::scene::Converter::getScale(pose->scale),
                     avi::scene::Converter::getRotation(pose->pose),
                     avi::scene::Converter::getColor(pose->color));
}

godot::Transform3D
avi::scene::PoseElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

void
avi::scene::PoseElementNode::setShape()
{
    DETOOLIFY;
    godot::Ref<godot::Shape3D> shape = mesh->create_trimesh_shape();
    this->collisionShape->set_shape(shape);
}

std::vector<godot::Node3D*>
avi::scene::PoseElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    vec.push_back(poseMeshInstance);
    return vec;
}
