#pragma once

#include "ElementNode.h"

#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/font.hpp>
#include <godot_cpp/classes/label3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/shape3d.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

#include <arviz_godot/lib/avi/core/LanguageUtility.h>

namespace avi::scene
{
    class PoseElementNode : public avi::scene::ElementNode
    {
        GDCLASS(PoseElementNode, ElementNode)

    public:
        static void _bind_methods();

        PoseElementNode();
        ~PoseElementNode() override = default;

        static PoseElementNode* newPoseElementNode(avi::connection::SentElementType type,
                                                   std::string name,
                                                   avi::core::Layer* layer);

        static PoseElementNode* newPoseElementNode();

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        void setColor(godot::Color color);

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        void setShape();

        void setStartAttributes(avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color);

        static godot::Mesh* mesh;

    private:
        const real_t textDownscaling = 0.125;

        godot::MeshInstance3D* poseMeshInstance;
        godot::Label3D* label;

        godot::StaticBody3D* staticBody;
        godot::CollisionShape3D* collisionShape;

        bool shapeChecked = false;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
