#pragma once

#include <vector>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/viewport.hpp>

#include <arviz_godot/lib/avi/core/OptionFileReader.h>
#include <arviz_godot/lib/avi/core/OptionFileWriter.h>
#include <arviz_godot/lib/avi/scene/DrawStyle.h>

namespace avi::scene
{
    struct Profile
    {
        godot::Viewport::MSAA msaa = godot::Viewport::MSAA::MSAA_DISABLED;
        godot::Color backgroundColor = godot::Color(1.0, 1.0, 1.0, 1.0);
        double ambientLight = 0.5f;
        double directionalLight = 0.5f;
        bool cameraHeadlight = false;
        double headlightStrength = 1.0f;
        int64_t fov = 70.0f;
        avi::scene::DrawStyle drawStyle = avi::scene::DrawStyle::ORIGINAL;
        bool showCoordinateLines = false;

    public:
        const std::string& getName();

        void rename(const std::string& name);

        void writeToFile(avi::core::OptionFile& optionFile);

        static avi::scene::Profile readFromFile(avi::core::OptionFile& optionFile);

        template <typename T>
        std::vector<std::unique_ptr<avi::core::Option<T>>>
        getOptions()
        {
            std::vector<std::unique_ptr<avi::core::Option<T>>> options;

            options.ADD_OPTION(godot::Viewport::MSAA, msaa, "MSAA", "Number of samples used for Multisample Anti-Aliasing.");
            options.ADD_OPTION(godot::Color, backgroundColor, "Background Color", "Clear color when rendering the scene.");
            options.ADD_OPTION(double, ambientLight, "Ambient Light Energy", "How bright the ambient (no direction) light is. Use 0.5 as default.");
            options.ADD_OPTION(double, directionalLight, "Directional Light Energy", "How bright the directional light is. Use 0.5 as default.");
            options.ADD_OPTION(bool, cameraHeadlight, "Camera Headlight", "Whether to activate the camera headlight.");
            options.ADD_OPTION(double, headlightStrength, "Headlight Energy", "How bright the camera headlight is.");
            options.ADD_OPTION(int64_t, fov, "FoV", "The field of view, in degrees.");
            options.ADD_OPTION(avi::scene::DrawStyle, drawStyle, "Draw Style", "Use this to override the draw style of robots and objects.");
            options.ADD_OPTION(bool, showCoordinateLines, "Coordinate Lines", "Whether to show the coordinate system axes in the scene.");

            return options;
        }

    private:
        std::string name{};
    };
} // namespace avi::scene
