#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ProfileStorage.h"

#include <memory>
#include <vector>

std::string* p;

avi::scene::ProfileStorage::ProfileStorage(const core::SettingsProvider& settingsProvider) :
    settingsProvider(settingsProvider)
{
    std::string directory = settingsProvider.getProfileDirectory();

    for (const auto& entry : std::filesystem::directory_iterator(directory))
    {
        std::filesystem::path entryPath = entry.path();

        if (entryPath.extension() == ".opt")
        {
            auto file = avi::core::OptionFile::load(entryPath.string());

            if (file.has_value())
            {
                files[file->getName()] = *file;

                auto profile = avi::scene::Profile::readFromFile(*file);
                profiles[profile.getName()] = profile;
            }
        }
    }
}

std::optional<avi::scene::Profile>
avi::scene::ProfileStorage::retrieveDefault()
{
    return retrieve(settingsProvider.getDefaultProfileName());
}

avi::scene::Profile
avi::scene::ProfileStorage::createDefault()
{
    avi::scene::Profile profile;
    profile.rename(settingsProvider.getDefaultProfileName());

    return profile;
}

std::vector<std::string>
avi::scene::ProfileStorage::getNames()
{
    std::vector<std::string> names;

    for (const auto& [name, profile] : profiles)
    {
        names.push_back(name);
    }

    return names;
}

bool
avi::scene::ProfileStorage::has(const std::string& name)
{
    return profiles.find(name) != profiles.end();
}

std::optional<avi::scene::Profile>
avi::scene::ProfileStorage::retrieve(const std::string& name)
{
    auto found = profiles.find(name);

    if (found != profiles.end())
    {
        avi::scene::Profile profile = std::move(found->second);
        profiles.erase(found);
        return profile;
    }

    return std::nullopt;
}

void
avi::scene::ProfileStorage::put(avi::scene::Profile profile)
{
    DETOOLIFY;
    profiles[profile.getName()] = std::move(profile);
}

void
avi::scene::ProfileStorage::save(avi::scene::Profile& profile)
{
    DETOOLIFY;
    std::filesystem::path path =
        settingsProvider.getProfileDirectory() / (profile.getName() + ".opt");

    auto& file = getOptionFile(profile.getName());
    profile.writeToFile(file);
    file.save(path);
}

void
avi::scene::ProfileStorage::cleanup()
{
    DETOOLIFY;
    for (const auto& [name, file] : files)
    {
        std::filesystem::path path = settingsProvider.getProfileDirectory() / (name + ".opt");

        if (profiles.find(name) == profiles.end())
        {
            std::filesystem::remove(path);
        }
    }
}

avi::core::OptionFile&
avi::scene::ProfileStorage::getOptionFile(const std::string& name)
{
    auto found = files.find(name);
    if (found != files.end())
    {
        return files.at(name);
    }
    else
    {
        files[name] = avi::core::OptionFile::create(name);
        return files[name];
    }
}
