#pragma once

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/node3d.hpp>
#include <arviz_godot/lib/avi/scene/CameraControlScheme.h>
#include <arviz_godot/lib/avi/scene/VisualizedScene.h>

namespace avi::scene
{
    class ViewCamera;

    class TechnicalControlScheme : public godot::Node3D, public avi::scene::CameraControlScheme
    {
        GDCLASS(TechnicalControlScheme, Node3D)

    public:
        static void _bind_methods();

        TechnicalControlScheme();
        ~TechnicalControlScheme() override = default;

        void _ready() override;

        void _unhandled_input(godot::Ref<godot::InputEvent> const& event) override;

        void _process(double delta) override;

    public:
        void setup(avi::core::SettingsProvider& settings,
                   avi::scene::VisualizedScene& scene,
                   bool gravity);

        void takeOver(avi::scene::ViewCamera* viewCamera, godot::Transform3D head) override;

        godot::Transform3D release() override;

        void reset() override;

        void focus(godot::AABB box) override;

        void focus(godot::Vector3 point, avi::core::Element* target) override;

        godot::Basis getRotationBasis() override;

        void placeOnAxis(godot::Vector3 fromVector) override;

        void invertAxis(godot::Vector3 fromVector) override;

        void setupControlAdvertisements(avi::core::ControlAdvertiser& advertiser) override;

    private:
        bool isActive();

        void orbit(godot::Vector2 rotation, float deltaTime);

        void rotateFixed(RotationDirection direction, HotkeyRotation hotkey) override;

        void placeOnAxisNumpad(Axis axis, AxisSign sign) override;

        void move(godot::Vector2 movement, float deltaTime);

        void zoom(float delta, float deltaTime);

        static float getClampedZoom(float zoom);

        void zoomOrthographic(float delta, float deltaTime);

        void roll(float delta, float deltaTime);

        void startTravel(godot::Transform3D anchor, float zoom, float duration);

        void doTravel(float deltaTime);

    private:
        struct Travel
        {
            godot::Transform3D startTransform{};
            godot::Transform3D endTransform{};
            godot::Transform3D endTransformHead{};

            float startZoom{};
            float endZoom{};

            float duration{};
            float remainingTime{};
        };

        godot::Vector2 mouseDelta;
        float wheelDelta;
        float smoothWheelDelta;

        float cameraStartSize = 1.0f;
        float cameraEndSize = 1.0f;
        bool updateSize = false;

        float zoomValue;
        float zoomDistance;

        bool gravity = false;

        bool switchX = false;
        bool switchY = false;
        bool switchZ = false;

        bool rotatePressed = false;
        bool continuousEnabled = true;
        godot::Vector2 pressedRotation;

        godot::Node3D* anchor;
        godot::Node3D* head;
        godot::Node3D* zoomer;
        godot::Node3D* attachment;

        avi::scene::ViewCamera* viewCamera;

        avi::core::SettingsProvider* settings;
        avi::scene::VisualizedScene* scene;

        Travel travel;
        bool isTraveling{};

        core::ControlAdvertiser* advertiser = nullptr;
        size_t adID;
    };
} // namespace avi::scene
