#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "TextElementNode.h"

#include <godot_cpp/classes/box_shape3d.hpp>
#include <godot_cpp/classes/font.hpp>
#include <godot_cpp/classes/mesh_instance3d.hpp>
#include <godot_cpp/classes/resource_loader.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

godot::Font* avi::scene::TextElementNode::font = nullptr;

float avi::scene::TextElementNode::scalingFactor = 0.125f;

avi::scene::TextElementNode::TextElementNode()
{
    DETOOLIFY;
}

godot::AABB
avi::scene::TextElementNode::getVolume()
{
    godot::AABB AABB = label->get_aabb();
    AABB.set_size(AABB.get_size() * getScale() * scalingFactor);
    AABB.position = staticBody->get_global_transform().origin;

    return AABB;
}

avi::scene::TextElementNode*
avi::scene::TextElementNode::newTextElementNode(godot::Node3D* scene)
{
    if (font == nullptr)
    {
        // We keep the font alive in the scene using the KeepAlive node.

        godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
        font = godot::Ref<godot::Font>(loader->load("res://fonts/font.tres")).ptr();
    }

    TextElementNode* node = memnew(TextElementNode);

    node->staticBody = memnew(godot::StaticBody3D);
    node->add_child(node->staticBody);

    node->label = memnew(godot::Label3D);
    node->staticBody->add_child(node->label);
    node->label->set_font(font);
    node->label->set_pixel_size(scalingFactor);
    node->label->set_font_size(128);
    node->label->set_offset({0, -5.0f / scalingFactor});
    node->label->set_horizontal_alignment(godot::HORIZONTAL_ALIGNMENT_LEFT);
    node->label->set_vertical_alignment(godot::VERTICAL_ALIGNMENT_BOTTOM);

    node->collisionShape = memnew(godot::CollisionShape3D);
    node->staticBody->add_child(node->collisionShape);

    return node;
}

void
avi::scene::TextElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quaternion rotation,
                                              godot::Color color,
                                              std::string text)
{
    DETOOLIFY;
    setInteractionFlag(flag);

    if (!shapeChecked)
    {
        if (ElementNode::isSelectable())
        {
            setShape();
        }
        shapeChecked = true;
    }

    setInteractions(interactions);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
    setText(text);
    setPosition(position);
}

void
avi::scene::TextElementNode::setStartAttributes(avi::connection::SentElementType type,
                                                std::string name,
                                                avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::TextElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);

    staticBody->set_transform(transform);
}

void
avi::scene::TextElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::TextElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::TextElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    label->set_modulate(color);
}

void
avi::scene::TextElementNode::setText(const std::string& newText)
{
    DETOOLIFY;
    if (text == newText)
    {
        return;
    }

    text = newText;

    label->set_text(text.c_str());

    collisionShape->set_scale(label->get_scale());
}

godot::Vector3
avi::scene::TextElementNode::getPosition()
{
    return staticBody->get_position() * 2;
}

godot::Vector3
avi::scene::TextElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::TextElementNode::getScale()
{
    return staticBody->get_scale();
}

void
avi::scene::TextElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto text = IceInternal::Handle<armarx::viz::data::ElementText>::dynamicCast(elementInfo);

    updateAttributes(text->interaction.enableFlags,
                     text->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(text->pose),
                     avi::scene::Converter::getScale(text->scale),
                     avi::scene::Converter::getRotation(text->pose),
                     avi::scene::Converter::getColor(text->color),
                     text->text);
}

godot::Transform3D
avi::scene::TextElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    return transform;
}

void
avi::scene::TextElementNode::setShape()
{
    DETOOLIFY;
    godot::Ref<godot::Shape3D> boxShape = memnew(godot::BoxShape3D);
    this->collisionShape->set_shape(boxShape);
}

std::vector<godot::Node3D*>
avi::scene::TextElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    vec.push_back(label);
    return vec;
}
