#pragma once

#include "ElementNode.h"
#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/font.hpp>
#include <godot_cpp/classes/label3d.hpp>
#include <godot_cpp/classes/packed_scene.hpp>
#include <godot_cpp/classes/resource_loader.hpp>
#include <godot_cpp/classes/viewport_texture.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

#include <arviz_godot/lib/avi/core/LanguageUtility.h>

namespace avi::scene
{
    class TextElementNode : public avi::scene::ElementNode
    {
        GDCLASS(TextElementNode, ElementNode)

    public:
        static TextElementNode* newTextElementNode(godot::Node3D* scene);

        TextElementNode();
        ~TextElementNode() override = default;

        godot::AABB getVolume() override;
        godot::Transform3D getTransform() override;

        void update(IceInternal::Handle<armarx::viz::data::Element> elementInfo) override;

        void setPosition(godot::Vector3 position);

        void setScale(godot::Vector3 scale);

        void setRotation(godot::Quaternion rotation);

        void setColor(godot::Color color);

        void setText(const std::string& newText);

        void setShape();

        godot::Vector3 getPosition();

        godot::Vector3 getRotation();

        godot::Vector3 getScale();

        void updateAttributes(int flag,
                              std::vector<std::string> interactions,
                              godot::Vector3 position,
                              godot::Vector3 scale,
                              godot::Quaternion rotation,
                              godot::Color color,
                              std::string text);

        void setStartAttributes(avi::connection::SentElementType type,
                                std::string name,
                                avi::core::Layer* layer);

    private:
        static godot::Font* font;

        godot::StaticBody3D* staticBody;
        godot::CollisionShape3D* collisionShape;

        godot::Label3D* label;

        static float scalingFactor;

        std::string text;

        bool shapeChecked = false;

    protected:
        std::vector<godot::Node3D*> getSpatial() override;
    };
} // namespace avi::scene
