#pragma once

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/physics_direct_space_state3d.hpp>
#include <godot_cpp/classes/camera3d.hpp>
#include <godot_cpp/variant/dictionary.hpp>

#include <arviz_godot/lib/avi/core/Camera.h>
#include <arviz_godot/lib/avi/core/ControlAdvertiser.h>
#include <arviz_godot/lib/avi/core/EventSubscriptionSystem.h>
#include <arviz_godot/lib/avi/core/SettingsProvider.h>
#include <arviz_godot/lib/avi/scene/GraphicsController.h>

namespace avi::scene
{
    class TechnicalControlScheme;
    class PlayfulControlScheme;
    class CameraControlScheme;
    class VisualizedScene;

    class ViewCamera : public godot::Camera3D, public avi::core::Camera
    {
        GDCLASS(ViewCamera, godot::Camera3D)

    public:
        static void _bind_methods();

        ViewCamera() = default;
        ~ViewCamera() override = default;

    public:
        void setup(avi::scene::VisualizedScene& scene,
                   avi::core::EventSubscriptionSystem& events,
                   avi::core::SettingsProvider& settings,
                   avi::scene::GraphicsController& graphics);

        void setupControlAdvertisements(avi::core::ControlAdvertiser& advertiser);

        void activate() override;

        avi::core::ScreenContent getCurrentScreenContent() override;

        godot::Vector2 getScreenSize() override;

        std::optional<std::pair<avi::core::Element*, godot::Node*>>
        getSelectedElement(godot::Vector2 point) override;

        std::optional<std::pair<avi::core::Element*, godot::Node*>>
        getSelectedElement(godot::Vector2 point, bool& anyHit) override;

        bool isMouseLocked();

        CameraControlScheme* getCurrentScheme();

        void visualizationChanged();

        void setClipPlanes(float near);

        bool isOrthogonal();

        void placeOnAxis(godot::Vector3 fromVector);

        void invertAxis(godot::Vector3 fromVector);

    private:
        void toggleLock();

        void focusPoint();

        void reset();

        void focusScene();

        void switchScheme();

        void switchProjection();

        void setMouseLocked(bool locked);

        std::optional<godot::Vector3> screenPointToWorldPoint(godot::Vector2 point);

    private:
        bool mouseLocked{false};
        bool orthogonal = false;

        avi::scene::TechnicalControlScheme* technicalControlScheme{nullptr};
        avi::scene::PlayfulControlScheme* playfulControlScheme{nullptr};
        avi::scene::CameraControlScheme* currentScheme{nullptr};

        avi::scene::VisualizedScene* scene{nullptr};

        godot::PhysicsDirectSpaceState3D* direct_space_state;
        godot::Vector3 ray_origin_position;
        godot::Vector3 ray_end_position;
        godot::Dictionary result_dict;

        avi::scene::GraphicsController* graphicsController;
    };
} // namespace avi::scene
