#pragma once

#include <mutex>
#include <vector>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/node3d.hpp>

#include <arviz_godot/lib/avi/connection/RevisionBuffer.h>
#include <arviz_godot/lib/avi/scene/Converter.h>
#include <arviz_godot/lib/avi/scene/LayerNode.h>

namespace avi::scene
{
    class VisualizedScene : public godot::Node3D
    {
        GDCLASS(VisualizedScene, godot::Node3D)

    public:
        static void _bind_methods();

        VisualizedScene() = default;
        ~VisualizedScene() = default;

        void _ready() override;

        void _process(double delta) override;
        void _notification(int what);

        void setup(avi::connection::RevisionBuffer& revisionBuffer,
                   avi::scene::GraphicsController* graphicsController);

        std::vector<std::string> getComponents();

        std::vector<std::string> getLayers(const std::string& component);

        void
        setLayerVisibility(const std::string& component, const std::string& layer, bool isVisible);
        bool getLayerVisibility(const std::string& component, const std::string& layer);

        void setComponentVisibility(const std::string& component, bool isVisible);
        bool getComponentVisibility(const std::string& component);

        godot::AABB getSceneVolume();

        godot::Vector3 getOrientation();

        void setOrientation(const godot::Vector3& orientation);

        void clearScene();

    private:
        real_t clipDegrees(real_t degrees);

        void visualizationChanged();

    private:
        struct Component
        {
            std::map<std::string, avi::scene::LayerNode*> layers;
            bool isVisible;
        };

        std::map<LayerName, avi::scene::LayerNode*> layers{};
        std::map<std::string, Component> components{};

        std::unique_ptr<avi::scene::Converter> converter{};
        avi::scene::GraphicsController* graphicsController;

        godot::Node3D* coordinateSystem;
    };
} // namespace avi::scene
