#include "avi/core/Detoolifier.h"
#include "avi/app/KeyInputEventDispatch.h"
#include "avi/app/Main.h"
#include "avi/gui/GraphicsSettings.h"
#include "avi/gui/LayerSelection.h"
#include "avi/gui/ManipulationMenu.h"
#include "avi/gui/OrientationWindow.h"
#include "avi/gui/ReplayCockpit.h"
#include "avi/gui/SelectionMenu.h"
#include "avi/gui/SettingsWindow.h"
#include "avi/gui/StatusIcons.h"
#include "avi/gui/UserInterface.h"
#include "avi/interaction/InteractionDispatch.h"
#include "avi/interaction/Move.h"
#include "avi/interaction/MoveAxisGizmo.h"
#include "avi/interaction/None.h"
#include "avi/interaction/Rotate.h"
#include "avi/interaction/RotateAxisGizmo.h"
#include "avi/interaction/Scale.h"
#include "avi/interaction/ScaleAxisGizmo.h"
#include "avi/scene/ArrowCircleElementNode.h"
#include "avi/scene/ArrowElementNode.h"
#include "avi/scene/CubeElementNode.h"
#include "avi/scene/LayerNode.h"
#include "avi/scene/ObjectElementNode.h"
#include "avi/scene/OrientationIndicator.h"
#include "avi/scene/PlayfulControlScheme.h"
#include "avi/scene/PointCloudElementNode.h"
#include "avi/scene/PoseElementNode.h"
#include "avi/scene/RobotElementNode.h"
#include "avi/scene/SphereElementNode.h"
#include "avi/scene/TechnicalControlScheme.h"
#include "avi/scene/TextElementNode.h"
#include "avi/scene/ViewCamera.h"
#include "avi/scene/VisualizedScene.h"

#include <gdextension_interface.h>
#include <godot_cpp/core/defs.hpp>
#include <godot_cpp/core/class_db.hpp>
#include <godot_cpp/godot.hpp>

#include <arviz_godot/lib/avi/gui/CameraModeMenu.h>
#include <arviz_godot/lib/avi/scene/CustomMeshElementNode.h>
#include <arviz_godot/lib/avi/scene/CylinderElementNode.h>
#include <arviz_godot/lib/avi/scene/CylindroidElementNode.h>
#include <arviz_godot/lib/avi/scene/EllipsoidElementNode.h>
#include <arviz_godot/lib/avi/scene/LineElementNode.h>
#include <arviz_godot/lib/avi/scene/PathElementNode.h>
#include <arviz_godot/lib/avi/scene/PolygonElementNode.h>

void register_types(godot::ModuleInitializationLevel p_level)
{
    if (p_level != godot::MODULE_INITIALIZATION_LEVEL_SCENE) {
        return;
    }

    godot::ClassDB::register_class<avi::app::KeyInputEventDispatch>();
    godot::ClassDB::register_class<avi::app::Main>();

    godot::ClassDB::register_class<avi::gui::GraphicsSettings>();
    godot::ClassDB::register_class<avi::gui::LayerSelection>();
    godot::ClassDB::register_class<avi::gui::ManipulationMenu>();
    godot::ClassDB::register_class<avi::gui::ReplayCockpit>();
    godot::ClassDB::register_class<avi::gui::SelectionMenu>();
    godot::ClassDB::register_class<avi::gui::SettingsWindow>();
    godot::ClassDB::register_class<avi::gui::StatusIcons>();
    godot::ClassDB::register_class<avi::gui::UserInterface>();
    godot::ClassDB::register_class<avi::gui::CameraModeMenu>();
    godot::ClassDB::register_class<avi::gui::OrientationWindow>();

    godot::ClassDB::register_class<avi::interaction::InteractionDispatch>();
    godot::ClassDB::register_class<avi::interaction::Move>();
    godot::ClassDB::register_class<avi::interaction::Rotate>();
    godot::ClassDB::register_class<avi::interaction::Scale>();
    godot::ClassDB::register_class<avi::interaction::None>();
    godot::ClassDB::register_class<avi::interaction::MoveAxisGizmo>();
    godot::ClassDB::register_class<avi::interaction::RotateAxisGizmo>();
    godot::ClassDB::register_class<avi::interaction::ScaleAxisGizmo>();

    godot::ClassDB::register_class<avi::scene::LayerNode>();
    godot::ClassDB::register_class<avi::scene::ElementNode>();
    godot::ClassDB::register_class<avi::scene::ObjectElementNode>();
    godot::ClassDB::register_class<avi::scene::PlayfulControlScheme>();
    godot::ClassDB::register_class<avi::scene::RobotElementNode>();
    godot::ClassDB::register_class<avi::scene::TechnicalControlScheme>();
    godot::ClassDB::register_class<avi::scene::TextElementNode>();
    godot::ClassDB::register_class<avi::scene::ArrowElementNode>();
    godot::ClassDB::register_class<avi::scene::ArrowCircleElementNode>();
    godot::ClassDB::register_class<avi::scene::PoseElementNode>();
    godot::ClassDB::register_class<avi::scene::ViewCamera>();
    godot::ClassDB::register_class<avi::scene::VisualizedScene>();
    godot::ClassDB::register_class<avi::scene::PointCloudElementNode>();
    godot::ClassDB::register_class<avi::scene::OrientationIndicator>();
    godot::ClassDB::register_abstract_class<avi::scene::MeshElementNode>();
    godot::ClassDB::register_class<avi::scene::CubeElementNode>();
    godot::ClassDB::register_class<avi::scene::SphereElementNode>();
    godot::ClassDB::register_class<avi::scene::CylinderElementNode>();
    godot::ClassDB::register_class<avi::scene::EllipsoidElementNode>();
    godot::ClassDB::register_class<avi::scene::CylindroidElementNode>();
    godot::ClassDB::register_class<avi::scene::LineElementNode>();
    godot::ClassDB::register_class<avi::scene::PathElementNode>();
    godot::ClassDB::register_class<avi::scene::CustomMeshElementNode>();
    godot::ClassDB::register_abstract_class<avi::scene::MultiMeshElementNode>();
    godot::ClassDB::register_class<avi::scene::PolygonElementNode>();
}

void unregister_types(godot::ModuleInitializationLevel p_level)
{
    if (p_level != godot::MODULE_INITIALIZATION_LEVEL_SCENE) {
        return;
    }
}

extern "C" {
    GDExtensionBool GDE_EXPORT gdavi_library_init(GDExtensionInterfaceGetProcAddress p_get_proc_address, const GDExtensionClassLibraryPtr p_library, GDExtensionInitialization *r_initialization)
    {
        godot::GDExtensionBinding::InitObject init_obj(p_get_proc_address, p_library, r_initialization);

        init_obj.register_initializer(register_types);
        init_obj.register_terminator(unregister_types);
        init_obj.set_minimum_library_initialization_level(godot::MODULE_INITIALIZATION_LEVEL_SCENE);

        return init_obj.init();
    }
}

