#include <vector>

#include <godot_cpp/classes/node3d.hpp>
#include <arviz_godot/lib/avi/app/Main.h>
#include <arviz_godot/lib/avi/connection/RevisionBuffer.h>
#include <arviz_godot/lib/avi/scene/Converter.h>
#include <arviz_godot/lib/test/avi/doctest.h>

int
getElementCount(godot::Node* layer)
{
    int numElements = 0;
    for (int i = 0; i < layer->get_child_count(); i++)
    {
        if (!layer->get_child(i)->is_queued_for_deletion())
            numElements++;
    }
    return numElements;
}

bool
getElementCountofValidElemenet(godot::Node* layer, int node)
{
    godot::Node* n;
    for (int i = node; i < layer->get_child_count(); i++)
    {
        if (!layer->get_child(i)->is_queued_for_deletion())
        {
            n = layer->get_child(i);
        }
    }
    return getElementCount(n);
}

void
setStandardParameter(IceUtil::Handle<armarx::viz::data::Element> element,
                     std::basic_string<char> id)
{
    element->scale = {1, 1, 1};
    element->color = {255, 255, 0, 0};
    element->interaction.contextMenuOptions = {id,
                                               "flags:",
                                               "SELECT",
                                               "CONTEXT_MENU",
                                               "TRANSLATION_X",
                                               "implicit global manipulation and transform show"};
    element->interaction.enableFlags = element->interaction.enableFlags |
                                       armarx::viz::data::InteractionEnableFlags::SELECT |
                                       armarx::viz::data::InteractionEnableFlags::CONTEXT_MENU |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_X |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_Y |
                                       armarx::viz::data::InteractionEnableFlags::TRANSLATION_Z |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_X |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_Y |
                                       armarx::viz::data::InteractionEnableFlags::SCALING_Z |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_X |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_Y |
                                       armarx::viz::data::InteractionEnableFlags::ROTATION_Z;
}

TEST_CASE("test converter: add and delete layer")
{
    // Actual tests should create temporary node trees for testing instead of using the existing scene tree.
    avi::app::Main* m = avi::app::Main::get();
    CHECK(m != nullptr);
    godot::Node3D* spatial = memnew(godot::Node3D);
    m->add_child(spatial);
    avi::connection::RevisionBuffer buffer;
    auto graphicsController = new avi::scene::GraphicsController(m);
    avi::scene::Converter converter{spatial, buffer, graphicsController};

    armarx::viz::data::LayerUpdates layerUpdates;
    armarx::viz::data::LayerUpdates layerUpdates2;

    armarx::viz::data::LayerUpdate layerUpdate;
    layerUpdate.name = "Layer 1";
    layerUpdate.action = armarx::viz::data::Layer_CREATE_OR_UPDATE;

    IceUtil::Handle<armarx::viz::data::ElementBox> element(new armarx::viz::data::ElementBox());
    element->id = "Box";
    element->size = {100, 100, 100};
    element->pose = {0, 0, 0, 0, 0, 0, 1};
    element->color = {255, 0, 0, 255};

    IceUtil::Handle<armarx::viz::data::ElementBox> elementLocalFlags(
        new armarx::viz::data::ElementBox());
    elementLocalFlags->id = "BoxLocal";
    elementLocalFlags->size = {100, 100, 100};
    elementLocalFlags->pose = {500, 0, 0, 0, 0, 0, 1};
    elementLocalFlags->interaction.enableFlags =
        armarx::viz::data::InteractionEnableFlags::SELECT |
        armarx::viz::data::InteractionEnableFlags::CONTEXT_MENU |
        armarx::viz::data::InteractionEnableFlags::ROTATION_LOCAL |
        armarx::viz::data::InteractionEnableFlags::TRANSLATION_LOCAL |
        armarx::viz::data::InteractionEnableFlags::SCALING_LOCAL;
    elementLocalFlags->interaction.contextMenuOptions = {elementLocalFlags->id,
                                                         "flags:",
                                                         "SELECT",
                                                         "CONTEXT_MENU",
                                                         "ROTATION_LOCAL",
                                                         "TRANSLATION_LOCAL",
                                                         "SCALING_LOCAL",
                                                         "this element uses local manipulation"};

    IceUtil::Handle<armarx::viz::data::ElementBox> elementTransformHide(
        new armarx::viz::data::ElementBox());
    elementTransformHide->id = "BoxTransformHide";
    elementTransformHide->size = {100, 100, 100};
    elementTransformHide->pose = {-500, 0, 0, 0, 0, 0, 1};
    elementTransformHide->interaction.enableFlags =
        armarx::viz::data::InteractionEnableFlags::TRANSFORM_HIDE |
        armarx::viz::data::InteractionEnableFlags::ROTATION_LOCAL |
        armarx::viz::data::InteractionEnableFlags::TRANSLATION_LOCAL |
        armarx::viz::data::InteractionEnableFlags::SCALING_LOCAL;
    elementTransformHide->interaction.contextMenuOptions = {
        elementTransformHide->id,
        "flags:",
        "SELECT",
        "CONTEXT_MENU",
        "TRANSFORM_HIDE",
        "when manipulating this element should only show the ghost"};
    elementTransformHide->color = {255, 0, 255, 0};

    IceUtil::Handle<armarx::viz::data::ElementSphere> sphere(
        new armarx::viz::data::ElementSphere());
    sphere->id = "Sphere";
    sphere->radius = 100;
    sphere->pose = {0, 500, 0, 0, 0, 0, 1};

    IceUtil::Handle<armarx::viz::data::ElementArrow> elementArrow(
        new armarx::viz::data::ElementArrow);
    elementArrow->id = "elementArrow";
    elementArrow->length = 250;
    elementArrow->pose = {500, 500, 0, 0, 0, 0, 1};

    IceUtil::Handle<armarx::viz::data::ElementArrowCircle> elementArrowCircle(
        new armarx::viz::data::ElementArrowCircle);
    elementArrowCircle->id = "elementArrowCircle";
    elementArrowCircle->radius = 100;
    elementArrowCircle->completion = 0.6;
    elementArrowCircle->pose = {-500, 500, 0, 0, 0, 0, 1};

    IceUtil::Handle<armarx::viz::data::ElementCylinder> elementCylinder(
        new armarx::viz::data::ElementCylinder);
    elementCylinder->id = "elementCylinder";
    elementCylinder->radius = 100;
    elementCylinder->height = 200;
    elementCylinder->pose = {-500, -500, 0, 0, 0, 0, 1};

    IceUtil::Handle<armarx::viz::data::ElementCylindroid> elementCylindroid(
        new armarx::viz::data::ElementCylindroid);
    elementCylindroid->id = "elementCylindroid";
    elementCylindroid->axisLengths = {100, 200}; //maybe not correct
    elementCylindroid->height = 100;
    elementCylindroid->pose = {500, -500, 0, 0, 0, 0, 1};
    elementCylindroid->scale = {100, 1, 50};

    IceUtil::Handle<armarx::viz::data::ElementEllipsoid> elementEllipsoid(
        new armarx::viz::data::ElementEllipsoid);
    elementEllipsoid->id = "elementEllipsoid";
    elementEllipsoid->axisLengths = {100, 150, 200};
    elementEllipsoid->pose = {0, -500, 0, 0, 0, 0, 1};
    layerUpdate.elements.push_back(element);
    layerUpdate.elements.push_back(elementLocalFlags);
    layerUpdate.elements.push_back(elementTransformHide);
    layerUpdate.elements.push_back(sphere);
    layerUpdate.elements.push_back(elementArrow);
    layerUpdate.elements.push_back(elementArrowCircle);
    layerUpdate.elements.push_back(elementCylinder);
    layerUpdate.elements.push_back(elementCylindroid);
    layerUpdate.elements.push_back(elementEllipsoid);

    armarx::viz::data::LayerUpdate layerUpdate2;
    layerUpdate2.name = "Layer 2";
    layerUpdate2.action = armarx::viz::data::Layer_CREATE_OR_UPDATE;

    layerUpdates.updates.emplace_back(layerUpdate);
    layerUpdates.updates.emplace_back(layerUpdate2);
    buffer.addUpdates(std::move(layerUpdates));

    auto r = converter.apply();

    CHECK(buffer.getChangedLayers().empty());
    CHECK(spatial->get_child_count() == 2);
    for (auto r1 : r)
    {
        CHECK(r1.second != nullptr);
        avi::scene::LayerNode* layerNode = r1.second;

        if (layerNode->getName() == "Layer 1")
        {
            CHECK(getElementCount(layerNode) == 9);
        }
        else if (layerNode->getName() == "Layer 2")
        {
            CHECK(getElementCount(layerNode) == 0);
        }
    }

    SUBCASE("delete all elements of layer 1")
    {
        armarx::viz::data::LayerUpdates updateLayer;
        armarx::viz::data::LayerUpdate update;
        update.name = "Layer 1";
        update.elements.clear();
        updateLayer.updates.emplace_back(update);
        buffer.addUpdates(std::move(updateLayer));
        CHECK(buffer.getChangedLayers().empty() == false);
        CHECK(buffer.getChangedLayers().size() == 1);
        CHECK(buffer.getChangedLayers().contains({"", "Layer 1"}));
        converter.apply();

        CHECK(spatial->get_child_count() == 2);
        CHECK(getElementCount(spatial->get_child(0)) == 0);
        CHECK(getElementCount(spatial->get_child(1)) == 0);

        CHECK(buffer.getChangedLayers().empty());

        SUBCASE("add previous elements to layer 1")
        {
            armarx::viz::data::LayerUpdates previousUpdate;
            previousUpdate.updates.emplace_back(layerUpdate);
            CHECK(layerUpdate.elements.size() == 9);
            buffer.addUpdates(std::move(previousUpdate));
            CHECK(buffer.getChangedLayers().size() == 1);
            CHECK(buffer.getChangedLayers().contains({"", "Layer 1"}));
            converter.apply();

            CHECK(buffer.getChangedLayers().empty());
            CHECK(spatial->get_child_count() == 2);
            for (auto r1 : r)
            {
                CHECK(r1.second != nullptr);
                avi::scene::LayerNode* layerNode = r1.second;

                if (layerNode->getName() == "Layer 1")
                {
                    CHECK(getElementCount(layerNode) == 9);
                }
                else if (layerNode->getName() == "Layer 2")
                {
                    CHECK(getElementCount(layerNode) == 0);
                }
            }
        }

        SUBCASE("delete elements of Layer 1")
        {
            armarx::viz::data::LayerUpdates updateLayer5;
            armarx::viz::data::LayerUpdate update5;
            update5.name = "Layer 1";
            update5.elements.clear();
            updateLayer5.updates.emplace_back(update5);
            buffer.addUpdates(std::move(updateLayer5));
            CHECK(buffer.getChangedLayers().empty() == false);
            CHECK(buffer.getChangedLayers().size() == 1);
            CHECK(buffer.getChangedLayers().contains({"", "Layer 1"}));
            converter.apply();

            CHECK(spatial->get_child_count() == 2);
            CHECK(getElementCount(spatial->get_child(0)) == 0);
            CHECK(getElementCount(spatial->get_child(1)) == 0);

            CHECK(buffer.getChangedLayers().empty());
        }
    }

    SUBCASE("delete Layer 1")
    {
        armarx::viz::data::LayerUpdate deleteLayer1;
        deleteLayer1.name = "Layer 1";
        deleteLayer1.action = armarx::viz::data::LayerAction::Layer_DELETE;

        layerUpdates2.updates.emplace_back(deleteLayer1);
        buffer.addUpdates(std::move(layerUpdates2));
        converter.apply();
        CHECK(getElementCount(spatial) == 1);

        CHECK(getElementCountofValidElemenet(spatial, 0) == 0);

        armarx::viz::data::LayerUpdates emptyUpdate;

        buffer.addUpdates(std::move(emptyUpdate));
        converter.apply();

        CHECK(getElementCount(spatial) == 1);
        CHECK(getElementCountofValidElemenet(spatial, 0) == 0);
    }

    SUBCASE("add elements to layer")
    {
        armarx::viz::data::LayerUpdates updates;

        IceUtil::Handle<armarx::viz::data::ElementPose> elementPose(
            new armarx::viz::data::ElementPose);
        elementPose->id = "elementPose";
        elementPose->pose = {-1000, 0, 0, 0, 0, 0, 1};

        //Cant load modell
        IceUtil::Handle<armarx::viz::data::ElementRobot> elementRobot(
            new armarx::viz::data::ElementRobot);
        elementRobot->id = "elementRobot";
        elementRobot->pose = {0, 1000, 0, 0, 0, 0, 1};

        IceUtil::Handle<armarx::viz::data::ElementText> elementText(
            new armarx::viz::data::ElementText);
        elementText->id = "elementText";
        elementText->text = "Hello World";
        elementText->pose = {1000, -1000, 0, 0, 0, 0, 1};
        elementText->scale = {25, 25, 25};

        layerUpdate.elements.push_back(elementPose);
        layerUpdate.elements.push_back(elementRobot);
        layerUpdate.elements.push_back(elementText);

        updates.updates.emplace_back(layerUpdate);
        CHECK(buffer.getChangedLayers().empty());
        buffer.addUpdates(std::move(updates));
        converter.apply();
        CHECK(getElementCount(spatial) == 2);
        for (auto r1 : r)
        {
            CHECK(r1.second != nullptr);
            avi::scene::LayerNode* layerNode = r1.second;

            if (layerNode->getName() == "Layer 1")
            {
                CHECK(getElementCount(layerNode) == 12);
            }
            else if (layerNode->getName() == "Layer 2")
            {
                CHECK(getElementCount(layerNode) == 0);
            }
        }
        CHECK(buffer.getChangedLayers().empty() == true);
    }

    SUBCASE("delete non-existing layer")
    {
        armarx::viz::data::LayerUpdates u;
        armarx::viz::data::LayerUpdate l;
        l.action = armarx::viz::data::Layer_DELETE;
        l.elements.clear();
        u.updates.emplace_back(l);

        buffer.addUpdates(std::move(u));
        converter.apply();

        CHECK(getElementCount(spatial) == 2);
    }

    spatial->free();
}

TEST_CASE("test clear()")
{
    avi::connection::RevisionBuffer buffer;
    avi::app::Main* m = avi::app::Main::get();
    auto graphicsController = new avi::scene::GraphicsController(m);
    avi::scene::Converter converter{memnew(godot::Node3D), buffer, graphicsController};
    armarx::viz::data::LayerUpdates updates;
    armarx::viz::data::LayerUpdate l1;
    l1.name = "1";
    armarx::viz::data::LayerUpdate l2;
    l1.name = "2";
    armarx::viz::data::LayerUpdate l3;
    l3.name = "3";
    armarx::viz::data::LayerUpdate l4;
    l4.name = "4";

    updates.updates.emplace_back(l1);
    updates.updates.emplace_back(l2);
    updates.updates.emplace_back(l3);
    updates.updates.emplace_back(l4);

    CHECK(buffer.getChangedLayers().size() == 0);
    buffer.addUpdates(std::move(updates));
    CHECK(buffer.getChangedLayers().size() == 4);
    converter.clear();

    CHECK(buffer.getChangedLayers().size() == 0);
}

TEST_CASE("all elements")
{

    avi::app::Main* m = avi::app::Main::get();
    CHECK(m != nullptr);
    godot::Node3D* spatial = memnew(godot::Node3D);
    m->add_child(spatial);
    avi::connection::RevisionBuffer buffer;
    auto graphicsController = new avi::scene::GraphicsController(m);
    avi::scene::Converter converter{spatial, buffer, graphicsController};

    armarx::viz::data::LayerUpdates layerUpdates;

    armarx::viz::data::LayerUpdate layerUpdate;
    layerUpdate.name = "Layer 1";
    layerUpdate.action = armarx::viz::data::Layer_CREATE_OR_UPDATE;

    SUBCASE("arrow")
    {
        IceUtil::Handle<armarx::viz::data::ElementArrow> elementArrow(
            new armarx::viz::data::ElementArrow);
        elementArrow->id = "elementArrow";
        elementArrow->length = 250;
        elementArrow->pose = {500, 500, 0, 0, 0, 0, 1};
        setStandardParameter(elementArrow, elementArrow->id);
        layerUpdate.elements.push_back(elementArrow);
    }
    SUBCASE("arrowcircle")
    {
        IceUtil::Handle<armarx::viz::data::ElementArrowCircle> elementArrowCircle(
            new armarx::viz::data::ElementArrowCircle);
        elementArrowCircle->id = "elementArrowCircle";
        elementArrowCircle->radius = 100;
        elementArrowCircle->completion = 0.6;
        elementArrowCircle->pose = {-500, 500, 0, 0, 0, 0, 1};
        setStandardParameter(elementArrowCircle, elementArrowCircle->id);
        layerUpdate.elements.push_back(elementArrowCircle);
    }
    SUBCASE("box")
    {
        IceUtil::Handle<armarx::viz::data::ElementBox> element(new armarx::viz::data::ElementBox());
        element->id = "Box";
        element->size = {100, 100, 100};
        element->pose = {0, 0, 0, 0, 0, 0, 1};
        setStandardParameter(element, element->id);
        element->color = {255, 0, 0, 255};
        layerUpdate.elements.push_back(element);
    }
    SUBCASE("cylinder")
    {
        IceUtil::Handle<armarx::viz::data::ElementCylinder> elementCylinder(
            new armarx::viz::data::ElementCylinder);
        elementCylinder->id = "elementCylinder";
        elementCylinder->radius = 100;
        elementCylinder->height = 200;
        elementCylinder->pose = {-500, -500, 0, 0, 0, 0, 1};
        setStandardParameter(elementCylinder, elementCylinder->id);
        layerUpdate.elements.push_back(elementCylinder);
    }
    SUBCASE("cylindroid")
    {
        IceUtil::Handle<armarx::viz::data::ElementCylindroid> elementCylindroid(
            new armarx::viz::data::ElementCylindroid);
        elementCylindroid->id = "elementCylindroid";
        elementCylindroid->axisLengths = {100, 200}; //maybe not correct
        elementCylindroid->height = 100;
        elementCylindroid->pose = {500, -500, 0, 0, 0, 0, 1};
        setStandardParameter(elementCylindroid, elementCylindroid->id);
        elementCylindroid->scale = {100, 1, 50};
        layerUpdate.elements.push_back(elementCylindroid);
    }
    SUBCASE("ellipsoid")
    {
        IceUtil::Handle<armarx::viz::data::ElementEllipsoid> elementEllipsoid(
            new armarx::viz::data::ElementEllipsoid);
        elementEllipsoid->id = "elementEllipsoid";
        elementEllipsoid->axisLengths = {100, 150, 200};
        elementEllipsoid->pose = {0, -500, 0, 0, 0, 0, 1};
        setStandardParameter(elementEllipsoid, elementEllipsoid->id);
        layerUpdate.elements.push_back(elementEllipsoid);
    }
    SUBCASE("line")
    {
        IceUtil::Handle<armarx::viz::data::ElementLine> elementLine(
            new armarx::viz::data::ElementLine);
        elementLine->id = "elementLine";

        elementLine->lineWidth = 50;
        elementLine->from = {1000, 0, 1000};
        elementLine->to = {1000, 0, 0};
        elementLine->scale = {50, 50, 50};
        elementLine->pose = {0, 1000, 0, 0, 0, 0, 1};
        setStandardParameter(elementLine, elementLine->id);
        layerUpdate.elements.push_back(elementLine);
    }
    SUBCASE("mesh")
    {
        IceUtil::Handle<armarx::viz::data::ElementMesh> elementMesh(
            new armarx::viz::data::ElementMesh);
        elementMesh->id = "elementMesh";
        //elementMesh->vertices = {{0,0,0}, {1000,1000,1000},{-1000,1000,1000}};
        elementMesh->scale = {50, 50, 50};
        elementMesh->pose = {0, -500, 0, 0, 0, 0, 1};
        //setStandardParameter(elementMesh, elementMesh->id);
        layerUpdate.elements.push_back(elementMesh);
    }
    SUBCASE("object")
    {
        IceUtil::Handle<armarx::viz::data::ElementObject> elementObject(
            new armarx::viz::data::ElementObject);
        elementObject->id = " elementObject";
        elementObject->filename = ""; //TODO element path
        elementObject->scale = {50, 50, 50};
        elementObject->pose = {1500, 0, 0, 0, 0, 0, 1};
        setStandardParameter(elementObject, elementObject->id);
        layerUpdate.elements.push_back(elementObject);
    }
    SUBCASE("path")
    {
        IceUtil::Handle<armarx::viz::data::ElementPath> elementPath(
            new armarx::viz::data::ElementPath);
        elementPath->id = "elementPath";
        elementPath->lineWidth = 50;
        elementPath->points = {{0, 0, 0}, {1000, 1000, 1000}, {500, 500, -500}};
        elementPath->scale = {50, 50, 50};
        elementPath->pose = {0, -500, 0, 0, 0, 0, 1};
        setStandardParameter(elementPath, elementPath->id);
        layerUpdate.elements.push_back(elementPath);
    }
    SUBCASE("point_cloud")
    {
        IceUtil::Handle<armarx::viz::data::ElementPointCloud> elementPointCloud(
            new armarx::viz::data::ElementPointCloud);
        elementPointCloud->id = "elementPointCloud";
        elementPointCloud->pointSizeInPixels = 1;
        elementPointCloud->points = {0xf, 0x1, 0x2};
        elementPointCloud->pose = {0, 0, 500, 0, 0, 0, 1};
        setStandardParameter(elementPointCloud, elementPointCloud->id);
        layerUpdate.elements.push_back(elementPointCloud);
    }
    SUBCASE("polygon")
    {
        IceUtil::Handle<armarx::viz::data::ElementPolygon> elementPolygon(
            new armarx::viz::data::ElementPolygon);
        elementPolygon->id = "elementPolygon";
        elementPolygon->points = {{0, 0, 0}, {500, 500, 500}, {-500, 500, 500}, {-500, -500, -500}};
        elementPolygon->pose = {1500, 0, 0, 0, 0, 0, 1};
        setStandardParameter(elementPolygon, elementPolygon->id);
        layerUpdate.elements.push_back(elementPolygon);
    }
    SUBCASE("pose")
    {
        IceUtil::Handle<armarx::viz::data::ElementPose> elementPose(
            new armarx::viz::data::ElementPose);
        elementPose->id = "elementPose";
        elementPose->pose = {-1000, 0, 0, 0, 0, 0, 1};
        setStandardParameter(elementPose, elementPose->id);
        layerUpdate.elements.push_back(elementPose);
    }
    SUBCASE("robot")
    {
        IceUtil::Handle<armarx::viz::data::ElementRobot> elementRobot(
            new armarx::viz::data::ElementRobot);
        elementRobot->id = "elementRobot";
        elementRobot->pose = {0, 1000, 0, 0, 0, 0, 1};
        elementRobot->project = "armar6_rt";
        elementRobot->filename = "armar6_rt/robotmodel/Armar6-SH/Armar6-SH.xml";
        setStandardParameter(elementRobot, elementRobot->id);
        layerUpdate.elements.push_back(elementRobot);
    }
    SUBCASE("sphere")
    {
        IceUtil::Handle<armarx::viz::data::ElementSphere> sphere(
            new armarx::viz::data::ElementSphere());
        sphere->id = "Sphere";
        sphere->radius = 100;
        sphere->pose = {0, 500, 0, 0, 0, 0, 1};
        setStandardParameter(sphere, sphere->id);
        layerUpdate.elements.push_back(sphere);
    }
    SUBCASE("text")
    {
        IceUtil::Handle<armarx::viz::data::ElementText> elementText(
            new armarx::viz::data::ElementText);
        elementText->id = "elementText";
        elementText->text = "Hello World";
        elementText->pose = {1000, -1000, 0, 0, 0, 0, 1};
        setStandardParameter(elementText, elementText->id);
        elementText->scale = {25, 25, 25};
        layerUpdate.elements.push_back(elementText);
    }

    layerUpdates.updates.emplace_back(layerUpdate);
    buffer.addUpdates(std::move(layerUpdates));
    converter.apply();
    CHECK(getElementCount(spatial) == 1);
    if (getElementCount(spatial) > 0)
        CHECK(getElementCount(spatial->get_child(0)) == 1);
    spatial->queue_free();
}
