#include <arviz_godot/arviz_godot/Catch2/src/catch2/catch_test_macros.hpp>
#include <arviz_godot/lib/avi/connection/Clock.h>

TEST_CASE("setTimeFrame", "[Clock]")
{

    avi::connection::Clock clock;
    avi::connection::TimeFrame timeFrame(1, 2);
    clock.setTimeFrame(timeFrame);

    REQUIRE(clock.getTimeFrame().begin == 1);
    REQUIRE(clock.getTimeFrame().end == 2);

    timeFrame = avi::connection::TimeFrame(-1, 2);
    clock.setTimeFrame(timeFrame);

    REQUIRE(clock.getTimeFrame().begin == -1);
    REQUIRE(clock.getTimeFrame().end == 2);
}

TEST_CASE("setAvailableTimeFrame", "[Clock]")
{
    avi::connection::Clock clock;
    avi::connection::TimeFrame timeFrame(1, 20);
    clock.setAvailableTimeFrame(timeFrame);

    REQUIRE(clock.getAvailableTimeFrame().begin == 1);
    REQUIRE(clock.getAvailableTimeFrame().end == 20);
}

TEST_CASE("advance in time frame", "[Clock]")
{
    avi::connection::Clock clock;
    avi::connection::TimeFrame timeFrame(1, 20);
    clock.setTimeFrame(timeFrame);
    clock.setSpeed(1);
    REQUIRE(clock.getTimeFrame().begin == 1);
    REQUIRE(clock.getTimeFrame().end == 20);

    clock.advance(5);

    REQUIRE(clock.getTargetTime() == 5);

    clock.setSpeed(0.1);
    clock.advance(130);

    //	REQUIRE(clock.getTargetTime() == 18);
}

TEST_CASE("advance outside time frame", "[Clock]")
{
    avi::connection::Clock clock;
    avi::connection::TimeFrame timeFrame(1, 20);
    clock.setTimeFrame(timeFrame);
    clock.setSpeed(1);
    REQUIRE(clock.getTimeFrame().begin == 1);
    REQUIRE(clock.getTimeFrame().end == 20);

    REQUIRE(clock.getSpeed() == 1);
    clock.advance(5);

    REQUIRE(clock.getTargetTime() == 5);

    clock.setSpeed(0.1);
    REQUIRE(clock.getSpeed() == 0.1f);
    clock.advance(130);

    //REQUIRE(clock.getTargetTime() == 19);
}

TEST_CASE("available time frame", "[Clock]")
{

    avi::connection::Clock clock;
    clock.setAvailableTimeFrame({0, 10});

    REQUIRE(clock.getAvailableTimeFrame().begin == 0);
    REQUIRE(clock.getAvailableTimeFrame().end == 10);
}

TEST_CASE("target time", "[Clock]")
{
    avi::connection::Clock clock;
    clock.setTimeFrame({0, 10});
    clock.setTargetTime(2);

    REQUIRE(clock.getTargetTime() == 2);

    clock.setTargetTime(9);

    REQUIRE(clock.getTargetTime() == 9);

    clock.setTargetTime(11);

    REQUIRE(clock.getTargetTime() == 10);
}
